# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Calculates Block1DCoeffs for a time step."""
import functools

import jax
import jax.numpy as jnp
from torax._src import constants
from torax._src import jax_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.core_profiles import updaters
from torax._src.fvm import block_1d_coeffs
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.sources import source_models as source_models_lib
from torax._src.sources import source_profile_builders
from torax._src.sources import source_profiles as source_profiles_lib
from torax._src.transport_model import transport_model as transport_model_lib


# pylint: disable=invalid-name
class CoeffsCallback:
  """Calculates Block1DCoeffs for a state."""

  def __init__(
      self,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      transport_model: transport_model_lib.TransportModel,
      explicit_source_profiles: source_profiles_lib.SourceProfiles,
      source_models: source_models_lib.SourceModels,
      evolving_names: tuple[str, ...],
      pedestal_model: pedestal_model_lib.PedestalModel,
  ):
    self.static_runtime_params_slice = static_runtime_params_slice
    self.transport_model = transport_model
    self.explicit_source_profiles = explicit_source_profiles
    self.source_models = source_models
    self.evolving_names = evolving_names
    self.pedestal_model = pedestal_model

  def __call__(
      self,
      dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
      x: tuple[cell_variable.CellVariable, ...],
      allow_pereverzev: bool = False,
      # Checks if reduced calc_coeffs for explicit terms when theta_implicit=1
      # should be called
      explicit_call: bool = False,
  ) -> block_1d_coeffs.Block1DCoeffs:
    """Returns coefficients given a state x.

    Used to calculate the coefficients for the implicit or explicit components
    of the PDE system.

    Args:
      dynamic_runtime_params_slice: Runtime configuration parameters. These
        values are potentially time-dependent and should correspond to the time
        step of the state x.
      geo: The geometry of the system at this time step.
      core_profiles: The core profiles of the system at this time step.
      x: The state with cell-grid values of the evolving variables.
      allow_pereverzev: If True, then the coeffs are being called within a
        linear solver. Thus could be either the use_predictor_corrector solver
        or as part of calculating the initial guess for the nonlinear solver. In
        either case, we allow the inclusion of Pereverzev-Corrigan terms which
        aim to stabilize the linear solver when being used with highly nonlinear
        (stiff) transport coefficients. The nonlinear solver solves the system
        more rigorously and Pereverzev-Corrigan terms are not needed.
      explicit_call: If True, then if theta_implicit=1, only a reduced
        Block1DCoeffs is calculated since most explicit coefficients will not
        be used.

    Returns:
      coeffs: The diffusion, convection, etc. coefficients for this state.
    """

    # Update core_profiles with the subset of new values of evolving variables
    core_profiles = updaters.update_core_profiles_during_step(
        x,
        self.static_runtime_params_slice,
        dynamic_runtime_params_slice,
        geo,
        core_profiles,
        self.evolving_names,
    )
    if allow_pereverzev:
      use_pereverzev = self.static_runtime_params_slice.solver.use_pereverzev
    else:
      use_pereverzev = False

    return calc_coeffs(
        self.static_runtime_params_slice,
        dynamic_runtime_params_slice=dynamic_runtime_params_slice,
        geo=geo,
        core_profiles=core_profiles,
        transport_model=self.transport_model,
        explicit_source_profiles=self.explicit_source_profiles,
        source_models=self.source_models,
        evolving_names=self.evolving_names,
        use_pereverzev=use_pereverzev,
        explicit_call=explicit_call,
        pedestal_model=self.pedestal_model,
    )


def _calculate_pereverzev_flux(
    dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    pedestal_model_output: pedestal_model_lib.PedestalModelOutput,
) -> tuple[jax.Array, jax.Array, jax.Array, jax.Array, jax.Array, jax.Array]:
  """Adds Pereverzev-Corrigan flux to diffusion terms."""

  consts = constants.CONSTANTS
  true_n_e_face = (
      core_profiles.n_e.face_value()

      * dynamic_runtime_params_slice.numerics.density_reference
  )
  true_n_i_face = (
      core_profiles.n_i.face_value()
      * dynamic_runtime_params_slice.numerics.density_reference
  )

  geo_factor = jnp.concatenate(
      [jnp.ones(1), geo.g1_over_vpr_face[1:] / geo.g0_face[1:]]
  )

  chi_face_per_ion = (
      geo.g1_over_vpr_face
      * true_n_i_face
      * consts.keV2J
      * dynamic_runtime_params_slice.solver.chi_pereverzev
  )

  chi_face_per_el = (
      geo.g1_over_vpr_face
      * true_n_e_face
      * consts.keV2J
      * dynamic_runtime_params_slice.solver.chi_pereverzev
  )

  d_face_per_el = dynamic_runtime_params_slice.solver.D_pereverzev
  v_face_per_el = (
      core_profiles.n_e.face_grad()
      / core_profiles.n_e.face_value()
      * d_face_per_el
      * geo_factor
  )

  # remove Pereverzev flux from boundary region if pedestal model is on
  # (for PDE stability)
  chi_face_per_ion = jnp.where(
      geo.rho_face_norm > pedestal_model_output.rho_norm_ped_top,
      0.0,
      chi_face_per_ion,
  )
  chi_face_per_el = jnp.where(
      geo.rho_face_norm > pedestal_model_output.rho_norm_ped_top,
      0.0,
      chi_face_per_el,
  )

  # set heat convection terms to zero out Pereverzev-Corrigan heat diffusion
  v_heat_face_ion = (
      core_profiles.T_i.face_grad()
      / core_profiles.T_i.face_value()
      * chi_face_per_ion
  )
  v_heat_face_el = (
      core_profiles.T_e.face_grad()
      / core_profiles.T_e.face_value()
      * chi_face_per_el
  )

  d_face_per_el = jnp.where(
      geo.rho_face_norm > pedestal_model_output.rho_norm_ped_top,
      0.0,
      d_face_per_el * geo.g1_over_vpr_face,
  )

  v_face_per_el = jnp.where(
      geo.rho_face_norm > pedestal_model_output.rho_norm_ped_top,
      0.0,
      v_face_per_el * geo.g0_face,
  )

  chi_face_per_ion = chi_face_per_ion.at[0].set(chi_face_per_ion[1])
  chi_face_per_el = chi_face_per_el.at[0].set(chi_face_per_el[1])

  return (
      chi_face_per_ion,
      chi_face_per_el,
      v_heat_face_ion,
      v_heat_face_el,
      d_face_per_el,
      v_face_per_el,
  )


def calc_coeffs(
    static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
    dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    transport_model: transport_model_lib.TransportModel,
    explicit_source_profiles: source_profiles_lib.SourceProfiles,
    source_models: source_models_lib.SourceModels,
    pedestal_model: pedestal_model_lib.PedestalModel,
    evolving_names: tuple[str, ...],
    use_pereverzev: bool = False,
    explicit_call: bool = False,
) -> block_1d_coeffs.Block1DCoeffs:
  """Calculates Block1DCoeffs for the time step described by `core_profiles`.

  Args:
    static_runtime_params_slice: General input parameters which are fixed
      through a simulation run, and if changed, would trigger a recompile.
    dynamic_runtime_params_slice: General input parameters that can change from
      time step to time step or simulation run to run, and do so without
      triggering a recompile.
    geo: Geometry describing the torus.
    core_profiles: Core plasma profiles for this time step during this iteration
      of the solver. Depending on the type of solver being used, this may or
      may not be equal to the original plasma profiles at the beginning of the
      time step.
    transport_model: A TransportModel subclass, calculates transport coeffs.
    explicit_source_profiles: Precomputed explicit source profiles. These
      profiles either do not depend on the core profiles or depend on the
      original core profiles at the start of the time step, not the "live"
      updating core profiles. For sources that are implicit, their explicit
      profiles are set to all zeros.
    source_models: All TORAX source/sink functions that generate the explicit
      and implicit source profiles used as terms for the core profiles
      equations.
    pedestal_model: A PedestalModel subclass, calculates pedestal values.
    evolving_names: The names of the evolving variables in the order that their
      coefficients should be written to `coeffs`.
    use_pereverzev: Toggle whether to calculate Pereverzev terms
    explicit_call: If True, indicates that calc_coeffs is being called for the
      explicit component of the PDE. Then calculates a reduced Block1DCoeffs if
      theta_implicit=1. This saves computation for the default fully implicit
      implementation.

  Returns:
    coeffs: Block1DCoeffs containing the coefficients at this time step.
  """

  # If we are fully implicit and we are making a call for calc_coeffs for the
  # explicit components of the PDE, only return a cheaper reduced Block1DCoeffs
  if explicit_call and static_runtime_params_slice.solver.theta_implicit == 1.0:
    return _calc_coeffs_reduced(
        geo,
        core_profiles,
        evolving_names,
    )
  else:
    return _calc_coeffs_full(
        static_runtime_params_slice,
        dynamic_runtime_params_slice,
        geo,
        core_profiles,
        transport_model,
        explicit_source_profiles,
        source_models,
        pedestal_model,
        evolving_names,
        use_pereverzev,
    )


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'static_runtime_params_slice',
        'transport_model',
        'pedestal_model',
        'source_models',
        'evolving_names',
    ],
)
def _calc_coeffs_full(
    static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
    dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    transport_model: transport_model_lib.TransportModel,
    explicit_source_profiles: source_profiles_lib.SourceProfiles,
    source_models: source_models_lib.SourceModels,
    pedestal_model: pedestal_model_lib.PedestalModel,
    evolving_names: tuple[str, ...],
    use_pereverzev: bool = False,
) -> block_1d_coeffs.Block1DCoeffs:
  """Calculates Block1DCoeffs for the time step described by `core_profiles`.

  Args:
    static_runtime_params_slice: General input parameters which are fixed
      through a simulation run, and if changed, would trigger a recompile.
    dynamic_runtime_params_slice: General input parameters that can change from
      time step to time step or simulation run to run, and do so without
      triggering a recompile.
    geo: Geometry describing the torus.
    core_profiles: Core plasma profiles for this time step during this iteration
      of the solver. Depending on the type of solver being used, this may or
      may not be equal to the original plasma profiles at the beginning of the
      time step.
    transport_model: A TransportModel subclass, calculates transport coeffs.
    explicit_source_profiles: Precomputed explicit source profiles. These
      profiles either do not depend on the core profiles or depend on the
      original core profiles at the start of the time step, not the "live"
      updating core profiles. For sources that are implicit, their explicit
      profiles are set to all zeros.
    source_models: All TORAX source/sink functions that generate the explicit
      and implicit source profiles used as terms for the core profiles
      equations.
    pedestal_model: A PedestalModel subclass, calculates pedestal values.
    evolving_names: The names of the evolving variables in the order that their
      coefficients should be written to `coeffs`.
    use_pereverzev: Toggle whether to calculate Pereverzev terms

  Returns:
    coeffs: Block1DCoeffs containing the coefficients at this time step.
  """

  consts = constants.CONSTANTS

  pedestal_model_output = pedestal_model(
      dynamic_runtime_params_slice, geo, core_profiles
  )

  # Boolean mask for enforcing internal temperature boundary conditions to
  # model the pedestal.
  # If rho_norm_ped_top_idx is outside of bounds of the mesh, the pedestal is
  # not present and the mask is all False. This is what is used in the case that
  # set_pedestal is False.
  mask = (
      jnp.zeros_like(geo.rho, dtype=bool)
      .at[pedestal_model_output.rho_norm_ped_top_idx]
      .set(True)
  )

  conductivity = source_models.conductivity.calculate_conductivity(
      dynamic_runtime_params_slice, geo, core_profiles
  )

  # Calculate the implicit source profiles and combines with the explicit
  merged_source_profiles = source_profile_builders.build_source_profiles(
      source_models=source_models,
      dynamic_runtime_params_slice=dynamic_runtime_params_slice,
      static_runtime_params_slice=static_runtime_params_slice,
      geo=geo,
      core_profiles=core_profiles,
      explicit=False,
      explicit_source_profiles=explicit_source_profiles,
      conductivity=conductivity,
  )

  # psi source terms. Source matrix is zero for all psi sources
  source_mat_psi = jnp.zeros_like(geo.rho)

  # fill source vector based on both original and updated core profiles
  source_psi = merged_source_profiles.total_psi_sources(geo)

  true_n_e = (
      core_profiles.n_e.value
      * dynamic_runtime_params_slice.numerics.density_reference
  )
  true_n_i = (
      core_profiles.n_i.value
      * dynamic_runtime_params_slice.numerics.density_reference
  )

  true_n_e_face = (
      core_profiles.n_e.face_value()
      * dynamic_runtime_params_slice.numerics.density_reference
  )
  true_n_i_face = (
      core_profiles.n_i.face_value()
      * dynamic_runtime_params_slice.numerics.density_reference
  )

  # Transient term coefficient vector (has radial dependence through r, n)
  toc_T_i = (
      1.5
      * geo.vpr ** (-2.0 / 3.0)
      * consts.keV2J
      * dynamic_runtime_params_slice.numerics.density_reference
  )
  tic_T_i = core_profiles.n_i.value * geo.vpr ** (5.0 / 3.0)
  toc_T_e = (
      1.5
      * geo.vpr ** (-2.0 / 3.0)
      * consts.keV2J
      * dynamic_runtime_params_slice.numerics.density_reference
  )
  tic_T_e = core_profiles.n_e.value * geo.vpr ** (5.0 / 3.0)
  toc_psi = (
      1.0
      / dynamic_runtime_params_slice.numerics.resistivity_multiplier
      * geo.rho_norm
      * conductivity.sigma
      * consts.mu0
      * 16
      * jnp.pi**2
      * geo.Phi_b**2
      / geo.F**2
  )
  tic_psi = jnp.ones_like(toc_psi)
  toc_dens_el = jnp.ones_like(geo.vpr)
  tic_dens_el = geo.vpr

  # Diffusion term coefficients
  transport_coeffs = transport_model(
      dynamic_runtime_params_slice, geo, core_profiles, pedestal_model_output
  )
  chi_face_ion = transport_coeffs.chi_face_ion
  chi_face_el = transport_coeffs.chi_face_el
  d_face_el = transport_coeffs.d_face_el
  v_face_el = transport_coeffs.v_face_el
  d_face_psi = geo.g2g3_over_rhon_face

  if static_runtime_params_slice.evolve_density:
    if d_face_el is None or v_face_el is None:
      raise NotImplementedError(
          f'{type(transport_model)} does not support the density equation.'
      )

  # entire coefficient preceding dT/dr in heat transport equations
  full_chi_face_ion = (
      geo.g1_over_vpr_face * true_n_i_face * consts.keV2J * chi_face_ion
  )
  full_chi_face_el = (
      geo.g1_over_vpr_face * true_n_e_face * consts.keV2J * chi_face_el
  )

  # entire coefficient preceding dne/dr in particle equation
  full_d_face_el = geo.g1_over_vpr_face * d_face_el
  full_v_face_el = geo.g0_face * v_face_el

  # density source terms. Initialize source matrix to zero
  source_mat_nn = jnp.zeros_like(geo.rho)

  # density source vector based both on original and updated core profiles
  source_n_e = merged_source_profiles.total_sources('n_e', geo)

  source_n_e += (
      mask
      * dynamic_runtime_params_slice.numerics.adaptive_n_source_prefactor
      * pedestal_model_output.n_e_ped
  )
  source_mat_nn += -(
      mask * dynamic_runtime_params_slice.numerics.adaptive_n_source_prefactor
  )

  # Pereverzev-Corrigan correction for heat and particle transport
  # (deals with stiff nonlinearity of transport coefficients)
  # TODO(b/311653933) this forces us to include value 0
  # convection terms in discrete system, slowing compilation down by ~10%.
  # See if can improve with a different pattern.
  (
      chi_face_per_ion,
      chi_face_per_el,
      v_heat_face_ion,
      v_heat_face_el,
      d_face_per_el,
      v_face_per_el,
  ) = jax.lax.cond(
      use_pereverzev,
      lambda: _calculate_pereverzev_flux(
          dynamic_runtime_params_slice,
          geo,
          core_profiles,
          pedestal_model_output,
      ),
      lambda: tuple([jnp.zeros_like(geo.rho_face)] * 6),
  )

  full_chi_face_ion += chi_face_per_ion
  full_chi_face_el += chi_face_per_el
  full_d_face_el += d_face_per_el
  full_v_face_el += v_face_per_el

  # Add Phi_b_dot terms to heat transport convection
  v_heat_face_ion += (
      -3.0
      / 4.0
      * geo.Phi_b_dot
      / geo.Phi_b
      * geo.rho_face_norm
      * geo.vpr_face
      * true_n_i_face
      * consts.keV2J
  )

  v_heat_face_el += (
      -3.0
      / 4.0
      * geo.Phi_b_dot
      / geo.Phi_b
      * geo.rho_face_norm
      * geo.vpr_face
      * true_n_e_face
      * consts.keV2J
  )

  # Add Phi_b_dot terms to particle transport convection
  full_v_face_el += (
      -1.0 / 2.0 * geo.Phi_b_dot / geo.Phi_b * geo.rho_face_norm * geo.vpr_face
  )

  # Add Phi_b_dot terms to poloidal flux convection
  v_face_psi = (
      -8.0
      * jnp.pi**2
      * consts.mu0
      * geo.Phi_b_dot
      * geo.Phi_b
      * conductivity.sigma_face
      * geo.rho_face_norm**2
      / geo.F_face**2
  )

  # Fill heat transport equation sources. Initialize source matrices to zero

  source_i = merged_source_profiles.total_sources('T_i', geo)
  source_e = merged_source_profiles.total_sources('T_e', geo)

  # Add the Qei effects.
  qei = merged_source_profiles.qei
  source_mat_ii = qei.implicit_ii * geo.vpr
  source_i += qei.explicit_i * geo.vpr
  source_mat_ee = qei.implicit_ee * geo.vpr
  source_e += qei.explicit_e * geo.vpr
  source_mat_ie = qei.implicit_ie * geo.vpr
  source_mat_ei = qei.implicit_ei * geo.vpr

  # Pedestal
  source_i += (
      mask
      * dynamic_runtime_params_slice.numerics.adaptive_T_source_prefactor
      * pedestal_model_output.T_i_ped
  )
  source_e += (
      mask
      * dynamic_runtime_params_slice.numerics.adaptive_T_source_prefactor
      * pedestal_model_output.T_e_ped
  )

  source_mat_ii -= (
      mask * dynamic_runtime_params_slice.numerics.adaptive_T_source_prefactor
  )

  source_mat_ee -= (
      mask * dynamic_runtime_params_slice.numerics.adaptive_T_source_prefactor
  )

  # Add effective Phi_b_dot heat source terms

  # second derivative of volume profile with respect to r_norm
  vprpr_norm = jnp.gradient(geo.vpr, geo.rho_norm)

  source_i += (
      1.0
      / 2.0
      * vprpr_norm
      * geo.Phi_b_dot
      / geo.Phi_b
      * geo.rho_norm
      * true_n_i
      * core_profiles.T_i.value
      * consts.keV2J
  )

  source_e += (
      1.0
      / 2.0
      * vprpr_norm
      * geo.Phi_b_dot
      / geo.Phi_b
      * geo.rho_norm
      * true_n_e
      * core_profiles.T_e.value
      * consts.keV2J
  )

  # Add effective Phi_b_dot poloidal flux source term

  ddrnorm_sigma_rnorm2_over_f2 = jnp.gradient(
      conductivity.sigma * geo.rho_norm**2 / geo.F**2,
      geo.rho_norm,
  )

  source_psi += (
      -8.0
      * jnp.pi**2
      * consts.mu0
      * geo.Phi_b_dot
      * geo.Phi_b
      * ddrnorm_sigma_rnorm2_over_f2
  )

  # Build arguments to solver based on which variables are evolving
  var_to_toc = {
      'T_i': toc_T_i,
      'T_e': toc_T_e,
      'psi': toc_psi,
      'n_e': toc_dens_el,
  }
  var_to_tic = {
      'T_i': tic_T_i,
      'T_e': tic_T_e,
      'psi': tic_psi,
      'n_e': tic_dens_el,
  }
  transient_out_cell = tuple(var_to_toc[var] for var in evolving_names)
  transient_in_cell = tuple(var_to_tic[var] for var in evolving_names)

  var_to_d_face = {
      'T_i': full_chi_face_ion,
      'T_e': full_chi_face_el,
      'psi': d_face_psi,
      'n_e': full_d_face_el,
  }
  d_face = tuple(var_to_d_face[var] for var in evolving_names)

  var_to_v_face = {
      'T_i': v_heat_face_ion,
      'T_e': v_heat_face_el,
      'psi': v_face_psi,
      'n_e': full_v_face_el,
  }
  v_face = tuple(var_to_v_face.get(var) for var in evolving_names)

  # d maps (row var, col var) to the coefficient for that block of the matrix
  # (Can't use a descriptive name or the nested comprehension to build the
  # matrix gets too long)
  d = {
      ('T_i', 'T_i'): source_mat_ii,
      ('T_i', 'T_e'): source_mat_ie,
      ('T_e', 'T_i'): source_mat_ei,
      ('T_e', 'T_e'): source_mat_ee,
      ('n_e', 'n_e'): source_mat_nn,
      ('psi', 'psi'): source_mat_psi,
  }
  source_mat_cell = tuple(
      tuple(d.get((row_block, col_block)) for col_block in evolving_names)
      for row_block in evolving_names
  )

  var_to_source = {
      'T_i': source_i,
      'T_e': source_e,
      'psi': source_psi,
      'n_e': source_n_e,
  }
  source_cell = tuple(var_to_source.get(var) for var in evolving_names)

  coeffs = block_1d_coeffs.Block1DCoeffs(
      transient_out_cell=transient_out_cell,
      transient_in_cell=transient_in_cell,
      d_face=d_face,
      v_face=v_face,
      source_mat_cell=source_mat_cell,
      source_cell=source_cell,
      auxiliary_outputs=(
          merged_source_profiles,
          conductivity,
          transport_coeffs,
      ),
  )

  return coeffs


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'evolving_names',
    ],
)
def _calc_coeffs_reduced(
    geo: geometry.Geometry,
    core_profiles: state.CoreProfiles,
    evolving_names: tuple[str, ...],
) -> block_1d_coeffs.Block1DCoeffs:
  """Calculates only the transient_in_cell terms in Block1DCoeffs."""

  # Only transient_in_cell is used for explicit terms if theta_implicit=1
  tic_T_i = core_profiles.n_i.value * geo.vpr ** (5.0 / 3.0)
  tic_T_e = core_profiles.n_e.value * geo.vpr ** (5.0 / 3.0)
  tic_psi = jnp.ones_like(geo.vpr)
  tic_dens_el = geo.vpr

  var_to_tic = {
      'T_i': tic_T_i,
      'T_e': tic_T_e,
      'psi': tic_psi,
      'n_e': tic_dens_el,
  }
  transient_in_cell = tuple(var_to_tic[var] for var in evolving_names)

  coeffs = block_1d_coeffs.Block1DCoeffs(
      transient_in_cell=transient_in_cell,
  )
  return coeffs



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for registering new pydantic configs."""
from torax._src.sources import base
from torax._src.sources import bremsstrahlung_heat_sink as bremsstrahlung_heat_sink_lib
from torax._src.sources import cyclotron_radiation_heat_sink as cyclotron_radiation_heat_sink_lib
from torax._src.sources import electron_cyclotron_source as electron_cyclotron_source_lib
from torax._src.sources import fusion_heat_source as fusion_heat_source_lib
from torax._src.sources import gas_puff_source as gas_puff_source_lib
from torax._src.sources import generic_current_source as generic_current_source_lib
from torax._src.sources import generic_ion_el_heat_source as generic_ion_el_heat_source_lib
from torax._src.sources import generic_particle_source as generic_particle_source_lib
from torax._src.sources import ion_cyclotron_source as ion_cyclotron_source_lib
from torax._src.sources import ohmic_heat_source as ohmic_heat_source_lib
from torax._src.sources import pellet_source as pellet_source_lib
from torax._src.sources import pydantic_model as sources_pydantic_model
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_heat_sink as impurity_radiation_heat_sink_lib
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_mavrin_fit as impurity_radiation_mavrin_fit_lib
from torax._src.torax_pydantic import model_config


def _validate_source_model_config(
    source_model_config_class: type[base.SourceModelBase],
    source_name: str,
):
  """Validates that the source model config is valid."""
  if source_name in ('qei', 'j_bootstrap'):
    raise ValueError(
        'Cannot register a new source model config for the qei or j_bootstrap'
        ' sources.'
    )

  source_model_config = source_model_config_class()
  if not hasattr(source_model_config, 'model_name'):
    raise ValueError(
        'The source model config must have a model_name attribute.'
    )
  model_name: str = source_model_config.model_name

  match source_name:
    case bremsstrahlung_heat_sink_lib.BremsstrahlungHeatSink.SOURCE_NAME:
      default_model_name = (
          bremsstrahlung_heat_sink_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case (
        cyclotron_radiation_heat_sink_lib.CyclotronRadiationHeatSink.SOURCE_NAME
    ):
      default_model_name = (
          cyclotron_radiation_heat_sink_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case electron_cyclotron_source_lib.ElectronCyclotronSource.SOURCE_NAME:
      default_model_name = (
          electron_cyclotron_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case gas_puff_source_lib.GasPuffSource.SOURCE_NAME:
      default_model_name = (
          gas_puff_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case generic_particle_source_lib.GenericParticleSource.SOURCE_NAME:
      default_model_name = (
          generic_particle_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case pellet_source_lib.PelletSource.SOURCE_NAME:
      default_model_name = (
          pellet_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case fusion_heat_source_lib.FusionHeatSource.SOURCE_NAME:
      default_model_name = (
          fusion_heat_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case (
        generic_ion_el_heat_source_lib.GenericIonElectronHeatSource.SOURCE_NAME
    ):
      default_model_name = (
          generic_ion_el_heat_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case impurity_radiation_heat_sink_lib.ImpurityRadiationHeatSink.SOURCE_NAME:
      default_model_name = (
          impurity_radiation_mavrin_fit_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case ion_cyclotron_source_lib.IonCyclotronSource.SOURCE_NAME:
      default_model_name = (
          ion_cyclotron_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case ohmic_heat_source_lib.OhmicHeatSource.SOURCE_NAME:
      default_model_name = (
          ohmic_heat_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case generic_current_source_lib.GenericCurrentSource.SOURCE_NAME:
      default_model_name = (
          generic_current_source_lib.DEFAULT_MODEL_FUNCTION_NAME
      )
    case _:
      raise ValueError(f'The source name {source_name} is not supported.')

  if model_name == default_model_name:
    raise ValueError(
        f'The model function name {model_name} must be different from'
        f' the default model function name {default_model_name} for'
        f' the source {source_name}.'
    )


def register_source_model_config(
    source_model_config_class: type[base.SourceModelBase],
    source_name: str,
):
  """Update Pydantic schema to include a source model config.

  See torax.torax_pydantic.tests.register_config_test.py for an example of how
  to use this function and expected behavior.

  Args:
    source_model_config_class: The new source model config to register. This
      should be a subclass of SourceModelBase that implements the interface and
      has a unique `model_name`.
    source_name: The name of the source to register the model config against.
      This should be one of the fields in the Sources pydantic model. For the
      two "special" sources ("qei" and "j_bootstrap") registering a new
      implementation is not supported.
  """
  _validate_source_model_config(source_model_config_class, source_name)
  # Update the Sources pydantic model to be aware of the new config.
  sources_pydantic_model.Sources.model_fields[
      f'{source_name}'
  ].annotation |= source_model_config_class
  # Rebuild the pydantic schema for both the Sources and ToraxConfig models so
  # that uses of either will have access to the new config.
  sources_pydantic_model.Sources.model_rebuild(force=True)
  model_config.ToraxConfig.model_rebuild(force=True)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions to load a config from config file or directory."""

import importlib
import os
import pathlib
import sys
import types
import typing
from typing import Any, Literal, TypeAlias
from torax._src.plotting import plotruns_lib
from torax._src.torax_pydantic import model_config


ExampleConfig: TypeAlias = Literal[
    'basic_config', 'iterhybrid_predictor_corrector', 'iterhybrid_rampup'
]

ExamplePlotConfig: TypeAlias = Literal[
    'default_plot_config',
    'global_params_plot_config',
    'simple_plot_config',
    'sources_plot_config',
]


def torax_path() -> pathlib.Path:
  """Returns the absolute path to the Torax directory."""

  path = pathlib.Path(__file__).parent.parent.parent
  assert path.is_dir(), f'Path {path} is not a directory.'
  assert path.parts[-1] == 'torax', f'Path {path} is not a Torax directory.'
  return path


def example_config_paths() -> dict[ExampleConfig, pathlib.Path]:
  """Returns a tuple of example config paths."""

  example_dir = torax_path().joinpath('examples')
  assert example_dir.is_dir()

  def _get_path(path):
    path = example_dir.joinpath(path + '.py')
    assert path.is_file(), f'Path {path} to the example config does not exist.'
    return path

  return {path: _get_path(path) for path in typing.get_args(ExampleConfig)}


def example_plot_config_paths() -> dict[ExamplePlotConfig, pathlib.Path]:
  """Returns a tuple of example plot config paths."""

  example_dir = torax_path().joinpath('plotting', 'configs')
  assert example_dir.is_dir(), f'Path {example_dir} is not a directory.'

  def _get_path(path):
    path = example_dir.joinpath(path + '.py')
    assert path.is_file(), f'Path {path} to the example config does not exist.'
    return path

  return {path: _get_path(path) for path in typing.get_args(ExamplePlotConfig)}


# Taken from
# https://docs.python.org/3/library/importlib.html#importing-a-source-file-directly
def _import_from_path(
    module_name: str, file_path: pathlib.Path
) -> types.ModuleType:

  spec = importlib.util.spec_from_file_location(module_name, file_path)
  module = importlib.util.module_from_spec(spec)
  sys.modules[module_name] = module
  if module is None:
    raise ValueError(f'No loader found for module {module_name}.')
  else:
    spec.loader.exec_module(module)  # pytype: disable=attribute-error
  return module


def import_module(path: str | pathlib.Path) -> dict[str, Any]:
  """Import a Python module from a file path as a dictionary.

  Relative paths are first resolved relative to the working directory, and then
  relative to the Torax directory.

  Args:
    path: The absolute or relative path to the Python module. The path can be
      represented as a string or a `pathlib.Path` object.

  Returns:
    All variables in the module as a dictionary.
  """

  path = pathlib.Path(path) if isinstance(path, str) else path

  if not path.is_file():
    # path.is_file() will return True if either path is an absolute path and it
    # exists, or if it is a relative path and it exists relative to the working
    # directory. In this case, neither was successful, so check whether this
    # path exists relative to the Torax directory.
    new_path = torax_path().joinpath(path)
    if not new_path.is_file():
      raise ValueError(
          f'The file {path} could not be found. If it is a relative path, it'
          ' could not be resolved relative to the working directory'
          f' {os.getcwd()} or the Torax directory {torax_path()}.'
      )
    path = new_path

  # An arbitrary module name is needed to import the config file.
  arbitrary_module_name = '_torax_temp_config_import'
  module = _import_from_path(arbitrary_module_name, path)
  return vars(module)


def build_torax_config_from_file(
    path: str | pathlib.Path,
) -> model_config.ToraxConfig:
  """Returns a ToraxConfig object from a config file.

  The config file is either a Python file containing a `CONFIG` dictionary, or
  a JSON file obtained via `ToraxConfig(...).model_dump_json`.

  Args:
    path: The absolute path to the config file. The path can be represented as a
      string or a `pathlib.Path` object.
  """

  path = pathlib.Path(path) if isinstance(path, str) else path

  match path.suffix:
    case '.json':
      raise NotImplementedError(
          'Loading ToraxConfig from JSON is not implemented yet.'
      )
    case '.nc':
      raise NotImplementedError(
          'Loading ToraxConfig from serialized JSON is not implemented yet.'
      )
    case '.py':
      cfg = import_module(path)
      if 'CONFIG' not in cfg:
        raise ValueError(
            f'The file {str(path)} is an invalid Torax config file, as it does'
            ' not have a `CONFIG` variable defined.'
        )
      return model_config.ToraxConfig.from_dict(cfg['CONFIG'])
    case _:
      raise ValueError(f'Path {path} is not a valid Torax config file.')


def get_plot_config_from_file(
    path: str | pathlib.Path,
) -> plotruns_lib.FigureProperties:
  """Returns a FigureProperties object from a config file.

  The config file is a Python file with a `PLOT_CONFIG` variable.

  Args:
    path: The absolute path to the config file. If `None`, the default plot
      config file is used.

  Returns:
    A FigureProperties object.
  """

  path = pathlib.Path(path) if isinstance(path, str) else path

  cfg = import_module(path)
  if 'PLOT_CONFIG' not in cfg:
    raise ValueError(
        f'The file {str(path)} is an invalid Torax plot config file, as it does'
        ' not have a `PLOT_CONFIG` variable defined.'
    )
  return cfg['PLOT_CONFIG']



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Routines for calculating impurity charge states."""

from typing import Final, Mapping, Sequence
import immutabledict
from jax import numpy as jnp
import numpy as np
from torax._src import array_typing
from torax._src import constants
from torax._src.config import plasma_composition

# Polynomial fit coefficients from A. A. Mavrin (2018):
# Improved fits of coronal radiative cooling rates for high-temperature plasmas,
# Radiation Effects and Defects in Solids, 173:5-6, 388-398,
# DOI: 10.1080/10420150.2018.1462361
_MAVRIN_Z_COEFFS: Final[Mapping[str, array_typing.ArrayFloat]] = (
    immutabledict.immutabledict({
        'C': np.array([  # Carbon
            [-7.2007e00, -1.2217e01, -7.3521e00, -1.7632e00, 5.8588e00],
            [0.0000e00, 0.0000e00, 0.0000e00, 0.0000e00, 6.0000e00],
        ]),
        'N': np.array([  # Nitrogen
            [0.0000e00, 3.3818e00, 1.8861e00, 1.5668e-01, 6.9728e00],
            [0.0000e00, 0.0000e00, 0.0000e00, 0.0000e00, 7.0000e00],
        ]),
        'O': np.array([  # Oxygen
            [0.0000e00, -1.8560e01, -3.8664e01, -2.2093e01, 4.0451e00],
            [-4.3092e00, -4.6261e-01, -3.7050e-02, 8.0180e-02, 7.9878e00],
            [0.0000e00, 0.0000e00, 0.0000e00, 0.0000e00, 8.0000e00],
        ]),
        'Ne': np.array([  # Neon
            [-2.5303e01, -6.4696e01, -5.3631e01, -1.3242e01, 8.9737e00],
            [-7.0678e00, 3.6868e00, -8.0723e-01, 2.1413e-01, 9.9532e00],
            [0.0000e00, 0.0000e00, 0.0000e00, 0.0000e00, 1.0000e01],
        ]),
        'Ar': np.array([  # Argon
            [6.8717e00, -1.1595e01, -4.3776e01, -2.0781e01, 1.3171e01],
            [-4.8830e-02, 1.8455e00, 2.5023e00, 1.1413e00, 1.5986e01],
            [-5.9213e-01, 3.5667e00, -8.0048e00, 7.9986e00, 1.4948e01],
        ]),
        'Kr': np.array([  # Krypton
            [1.3630e02, 4.6320e02, 5.6890e02, 3.0638e02, 7.7040e01],
            [-1.0279e02, 6.8446e01, 1.5744e01, 1.5186e00, 2.4728e01],
            [-2.4682e00, 1.3215e01, -2.5703e01, 2.3443e01, 2.5368e01],
        ]),
        'Xe': np.array([  # Xenon
            [5.8178e02, 1.9967e03, 2.5189e03, 1.3973e03, 3.0532e02],
            [8.6824e01, -2.9061e01, -4.8384e01, 1.6271e01, 3.2616e01],
            [4.0756e02, -9.0008e02, 6.6739e02, -1.7259e02, 4.8066e01],
            [-1.0019e01, 7.3261e01, -1.9931e02, 2.4056e02, -5.7527e01],
        ]),
        'W': np.array([  # Tungsten
            [1.6823e01, 3.4582e01, 2.1027e01, 1.6518e01, 2.6703e01],
            [-2.5887e02, -1.0577e01, 2.5532e02, -7.9611e01, 3.6902e01],
            [1.5119e01, -8.4207e01, 1.5985e02, -1.0011e02, 6.3795e01],
        ]),
    })
)

# Temperature boundaries in keV, separating the rows for the fit coefficients.
_TEMPERATURE_INTERVALS: Final[Mapping[str, array_typing.ArrayFloat]] = (
    immutabledict.immutabledict({
        'C': np.array([0.7]),
        'N': np.array([0.7]),
        'O': np.array([0.3, 1.5]),
        'Ne': np.array([0.5, 2.0]),
        'Ar': np.array([0.6, 3.0]),
        'Kr': np.array([0.447, 4.117]),
        'Xe': np.array([0.3, 1.5, 8.0]),
        'W': np.array([1.5, 4.0]),
    })
)


# pylint: disable=invalid-name
def calculate_average_charge_state_single_species(
    T_e: array_typing.ArrayFloat,
    ion_symbol: str,
) -> array_typing.ArrayFloat:
  """Calculates the average charge state of an impurity based on the Marvin 2018 polynomial fit.

  The polynomial fit range is 0.1-100 keV, which is well within the typical
  bounds of core tokamak modelling. For safety, inputs are clipped to avoid
  extrapolation outside this range.

  Args:
    T_e: Electron temperature [keV].
    ion_symbol: Species to calculate average charge state for.

  Returns:
    Z: Average charge state [amu].
  """

  if ion_symbol not in constants.ION_SYMBOLS:
    raise ValueError(
        f'Invalid ion symbol: {ion_symbol}. Allowed symbols are :'
        f' {constants.ION_SYMBOLS}'
    )
  # Return the Z value for light ions that are fully ionized for T > 0.1 keV.
  if ion_symbol not in _MAVRIN_Z_COEFFS:
    return jnp.ones_like(T_e) * constants.ION_PROPERTIES_DICT[ion_symbol].Z

  # Avoid extrapolating fitted polynomial out of bounds.
  T_e_allowed_range = (0.1, 100.0)
  T_e = jnp.clip(T_e, *T_e_allowed_range)

  # Gather coefficients for each temperature
  interval_indices = jnp.searchsorted(_TEMPERATURE_INTERVALS[ion_symbol], T_e)
  Zavg_coeffs_in_range = jnp.take(
      _MAVRIN_Z_COEFFS[ion_symbol], interval_indices, axis=0
  ).transpose()

  # Calculate Zavg from the polynomial fit.
  X = jnp.log10(T_e)
  Zavg = jnp.polyval(Zavg_coeffs_in_range, X)

  return Zavg


def get_average_charge_state(
    ion_symbols: Sequence[str],
    ion_mixture: plasma_composition.DynamicIonMixture,
    T_e: array_typing.ArrayFloat,
) -> array_typing.ArrayFloat:
  """Calculates or prescribes average impurity charge state profile (JAX-compatible).

  Args:
    ion_symbols: Species to calculate average charge state for.
    ion_mixture: DynamicIonMixture object containing impurity information. The
      index of the ion_mixture.fractions array corresponds to the index of the
      ion_symbols array.
    T_e: Electron temperature [keV]. Can be any sized array, e.g. on cell grid,
      face grid, or a single scalar.

  Returns:
    avg_Z: Average charge state profile [amu].
      The shape of avg_Z is the same as T_e.
  """

  if ion_mixture.Z_override is not None:
    return jnp.ones_like(T_e) * ion_mixture.Z_override

  avg_Z = jnp.zeros_like(T_e)
  for ion_symbol, fraction in zip(ion_symbols, ion_mixture.fractions):
    avg_Z += fraction * calculate_average_charge_state_single_species(
        T_e, ion_symbol
    )

  return avg_Z



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Full simulator state to be used for orchestration."""
from absl import logging
import chex
import jax
import jax.numpy as jnp
from torax._src import state
from torax._src.geometry import geometry
from torax._src.sources import source_profiles


@chex.dataclass
class ToraxSimState:
  """Full simulator state.

  The simulation stepping in sim.py evolves core_profiles which includes all
  the attributes the simulation is advancing. But beyond those, there are
  additional stateful elements which may evolve on each simulation step, such
  as sources and transport.

  This class includes both core_profiles and these additional elements.

  Attributes:
    t: time coordinate.
    dt: timestep interval.
    core_profiles: Core plasma profiles at time t.
    core_transport: Core plasma transport coefficients computed at time t.
    core_sources: Profiles for all sources/sinks. These are the profiles that
      are used to calculate the coefficients for the t+dt time step. For the
      explicit sources, these are calculated at the start of the time step, so
      are the values at time t. For the implicit sources, these are the most
      recent guess for time t+dt. The profiles here are the merged version of
      the explicit and implicit profiles.
    geometry: Geometry at this time step used for the simulation.
    solver_numeric_outputs: Numerical quantities related to the solver.
  """

  t: jax.Array
  dt: jax.Array
  core_profiles: state.CoreProfiles
  core_transport: state.CoreTransport
  core_sources: source_profiles.SourceProfiles
  geometry: geometry.Geometry
  solver_numeric_outputs: state.SolverNumericOutputs

  def check_for_errors(self) -> state.SimError:
    """Checks for errors in the simulation state."""
    if self.core_profiles.negative_temperature_or_density():
      logging.info("%s", self.core_profiles)
      _log_negative_profile_names(self.core_profiles)
      return state.SimError.NEGATIVE_CORE_PROFILES
    # If there are NaNs that occurred without negative core profiles, log this
    # as a separate error.
    if _has_nan(self):
      logging.info("%s", self.core_profiles)
      return state.SimError.NAN_DETECTED
    elif not self.core_profiles.quasineutrality_satisfied():
      return state.SimError.QUASINEUTRALITY_BROKEN
    else:
      return state.SimError.NO_ERROR


def _has_nan(
    inputs: ToraxSimState,
) -> bool:
  return any([jnp.any(jnp.isnan(x)) for x in jax.tree.leaves(inputs)])


def _log_negative_profile_names(inputs: state.CoreProfiles):
  path_vals, _ = jax.tree.flatten_with_path(inputs)
  for path, value in path_vals:
    if jnp.any(jnp.less(value, 0.0)):
      logging.info("Found negative value in %s", jax.tree_util.keystr(path))



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Plasma composition parameters used throughout TORAX simulations."""
import functools

import chex
import numpy as np
from torax._src import array_typing
from torax._src import constants
from torax._src.config import runtime_validation_utils
from torax._src.torax_pydantic import torax_pydantic

# pylint: disable=invalid-name


@chex.dataclass(frozen=True)
class DynamicIonMixture:
  """Represents a fixed mixture of ion species at a specific time.

  Information on ion names are not stored here, but rather in
  StaticRuntimeParamsSlice, to simplify JAX logic and performance in source
  functions for fusion power and radiation which are species-dependent.

  Attributes:
    fractions: Ion fractions for a time slice.
    avg_A: Average A of the mixture.
    Z_override: Typically, the average Z is calculated according to the
      temperature dependent charge-state-distribution, or for low-Z cases by the
      atomic numbers of the ions assuming full ionization. If Z_override is
      provided, it is used instead for the average Z.
  """

  fractions: array_typing.ArrayFloat
  avg_A: array_typing.ScalarFloat
  Z_override: array_typing.ScalarFloat | None = None


class IonMixture(torax_pydantic.BaseModelFrozen):
  """Represents a mixture of ion species. The mixture can depend on time.

  Main use cases:
  1. Represent a bundled mixture of hydrogenic main ions (e.g. D and T)
  2. Represent a bundled impurity species where the avg charge state, mass,
    and radiation is consistent with each fractional concentration, and these
    quantities are then averaged over the mixture to represent a single impurity
    species in the transport equations for efficiency.

  Attributes:
    species: A dict mapping ion symbols (from ION_SYMBOLS) to their fractional
      concentration in the mixture. The fractions must sum to 1.
    Z_override: An optional override for the average charge (Z) of the mixture.
    A_override: An optional override for the average mass (A) of the mixture.
  """

  species: runtime_validation_utils.IonMapping
  Z_override: torax_pydantic.TimeVaryingScalar | None = None
  A_override: torax_pydantic.TimeVaryingScalar | None = None

  def build_dynamic_params(
      self,
      t: chex.Numeric,
  ) -> DynamicIonMixture:
    """Creates a DynamicIonMixture object at a given time.

    Optional overrides for Z and A can be provided.

    Args:
      t: The time at which to build the DynamicIonMixture.

    Returns:
      A DynamicIonMixture object.
    """
    ions = self.species.keys()
    fractions = np.array([self.species[ion].get_value(t) for ion in ions])
    Z_override = None if not self.Z_override else self.Z_override.get_value(t)

    if not self.A_override:
      As = np.array([constants.ION_PROPERTIES_DICT[ion].A for ion in ions])
      avg_A = np.sum(As * fractions)
    else:
      avg_A = self.A_override.get_value(t)

    return DynamicIonMixture(
        fractions=fractions,
        avg_A=avg_A,
        Z_override=Z_override,
    )


@chex.dataclass
class DynamicPlasmaComposition:
  main_ion: DynamicIonMixture
  impurity: DynamicIonMixture
  Z_eff: array_typing.ArrayFloat
  Z_eff_face: array_typing.ArrayFloat


class PlasmaComposition(torax_pydantic.BaseModelFrozen):
  """Configuration for the plasma composition.

  The `from_dict(...)` method can accept a dictionary defined by
  https://torax.readthedocs.io/en/latest/configuration.html#plasma-composition.

  Attributes:
    main_ion: Main ion species. Can be single ion or a mixture of ions (e.g. D
      and T). Either a single ion, and constant mixture, or a time-dependent
      mixture. For single ions the input is one of the allowed strings in
      `ION_SYMBOLS`. For mixtures the input is an IonMixture object, constructed
      from a dict mapping ion symbols to their fractional concentration in the
      mixture.
    impurity: Impurity ion species. Same format as main_ion.
    Z_eff: Constraint for impurity densities.
    Z_i_override: Optional arbitrary masses and charges which can be used to
      override the data for the average Z and A of each IonMixture for main_ions
      or impurities. Useful for testing or testing physical sensitivities,
      outside the constraint of allowed impurity species.
    A_i_override: Optional arbitrary masses and charges which can be used to
      override the data for the average Z and A of each IonMixture for main_ions
      or impurities. Useful for testing or testing physical sensitivities,
      outside the constraint of allowed impurity species.
    Z_impurity_override: Optional arbitrary masses and charges which can
  """

  main_ion: runtime_validation_utils.IonMapping = (
      torax_pydantic.ValidatedDefault({'D': 0.5, 'T': 0.5})
  )
  impurity: runtime_validation_utils.IonMapping = (
      torax_pydantic.ValidatedDefault('Ne')
  )
  Z_eff: (
      runtime_validation_utils.TimeVaryingArrayDefinedAtRightBoundaryAndBounded
  ) = torax_pydantic.ValidatedDefault(1.0)
  Z_i_override: torax_pydantic.TimeVaryingScalar | None = None
  A_i_override: torax_pydantic.TimeVaryingScalar | None = None
  Z_impurity_override: torax_pydantic.TimeVaryingScalar | None = None
  A_impurity_override: torax_pydantic.TimeVaryingScalar | None = None

  # Generate the IonMixture objects from the input for either a mixture (dict)
  # or the shortcut for a single ion (string). IonMixture objects with a
  # single key and fraction=1.0 is used also for the single ion case to reduce
  # code duplication.

  @functools.cached_property
  def main_ion_mixture(self) -> IonMixture:
    """Returns the IonMixture object for the main ions."""
    # Use `model_construct` as no validation required.
    return IonMixture.model_construct(
        species=self.main_ion,
        Z_override=self.Z_i_override,
        A_override=self.A_i_override,
    )

  @functools.cached_property
  def impurity_mixture(self) -> IonMixture:
    """Returns the IonMixture object for the impurity ions."""
    # Use `model_construct` as no validation required.
    return IonMixture.model_construct(
        species=self.impurity,
        Z_override=self.Z_impurity_override,
        A_override=self.A_impurity_override,
    )

  def get_main_ion_names(self) -> tuple[str, ...]:
    """Returns the main ion symbol strings from the input."""
    return tuple(self.main_ion_mixture.species.keys())

  def get_impurity_names(self) -> tuple[str, ...]:
    """Returns the impurity symbol strings from the input."""
    return tuple(self.impurity_mixture.species.keys())

  def build_dynamic_params(self, t: chex.Numeric) -> DynamicPlasmaComposition:
    return DynamicPlasmaComposition(
        main_ion=self.main_ion_mixture.build_dynamic_params(t),
        impurity=self.impurity_mixture.build_dynamic_params(t),
        Z_eff=self.Z_eff.get_value(t),
        Z_eff_face=self.Z_eff.get_value(t, grid_type='face'),
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=invalid-name

"""Cyclotron radiation heat sink for electron heat equation.."""
import dataclasses
from typing import ClassVar, Literal

import chex
import jax
from jax import numpy as jnp
import pydantic
from torax._src import array_typing
from torax._src import jax_utils
from torax._src import math_utils
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.sources import base
from torax._src.sources import runtime_params as runtime_params_lib
from torax._src.sources import source
from torax._src.sources import source_profiles
import typing_extensions


# Default value for the model function to be used for the Cyclotron radiation
# heat sink source. This is also used as an identifier for the model function in
# the default source config for Pydantic to "discriminate" against.
DEFAULT_MODEL_FUNCTION_NAME: str = 'albajar_artaud'


@chex.dataclass(frozen=True)
class StaticRuntimeParams(runtime_params_lib.StaticRuntimeParams):
  beta_min: float
  beta_max: float
  beta_grid_size: int


@chex.dataclass(frozen=True)
class DynamicRuntimeParams(runtime_params_lib.DynamicRuntimeParams):
  wall_reflection_coeff: array_typing.ScalarFloat


def _alpha_closed_form(
    *,
    beta: array_typing.ScalarFloat,
    rho_norm: array_typing.ArrayFloat,
    profile_data: array_typing.ArrayFloat,
    profile_edge_value: array_typing.ScalarFloat,
) -> array_typing.ScalarFloat:
  """Returns analytical closed form of alpha for parameterized profiles.

  See albajar_artaud for more details.

  We find alpha for the best fit for either the n_e_data or te_data to the
  parameterized functions:

  n_e = n_e_data[0]*(1 - rhonorm**2)**alpha
  te = (te_data[0] - te_edge)*(1 - rhonorm**beta)**alpha + te_edge

  Where n_e_edge = 0 above. The parameterizations are from the Albajar paper.

  The fit is from the magnetic axis to rhonorm=0.9, to avoid pedestal effects.

  We analytically solve for alpha by taking the derivative of the loss function
  with respect to alpha.

  Defining:
  T = (profile_data - profile_edge_data)/(profile_data[0] - profile_edge_data)

  The loss function is:
  loss = sum((log(T) -  alpha * log(1 - rhonorm**beta)) ** 2) / 2

  The solution is provided by setting d loss / d alpha = 0.

  Args:
    beta: The beta parameter to use in the parameterized functions. For the
      density fit, this is always 2.
    rho_norm: Normalized toroidal flux coordinate.
    profile_data: The profile data to be fit (i.e. density or temperature).
    profile_edge_value: The value of profile data at the edge to use. This can't
      be taken from profile_data since there are different assumptions for the
      temperature and density fits.

  Returns:
    The alpha parameter being fit for either the density or temperature fits.
  """
  # To avoid dealing with slicing of non-concrete values, we do a masking trick
  # where we replace the values of n_e_data and rhonorm above 0.9 with values
  # that will not contribute to the sums in the numerator and denominator.

  mask = rho_norm < 0.9
  profile_data_norm = (profile_data - profile_edge_value) / (
      profile_data[0] - profile_edge_value
  )
  sliced_profile_data_norm = jnp.where(mask, profile_data_norm, 1.0)
  sliced_rhonorm = jnp.where(mask, rho_norm, 0.0)

  num = jnp.sum(
      (jnp.log(sliced_profile_data_norm) * jnp.log(1 - sliced_rhonorm**beta))
  )
  den = jnp.sum(jnp.log(1 - sliced_rhonorm**beta) ** 2)

  alpha_n = num / den
  return alpha_n


def _loss_for_beta_t(
    beta_t: array_typing.ScalarFloat,
    rho_norm: array_typing.ArrayFloat,
    te_data: array_typing.ArrayFloat,
) -> array_typing.ScalarFloat:
  """Returns the loss function for the temperature fit for a given beta_t.

  The fit is from the magnetic axis to rhonorm=0.9, to avoid pedestal effects.
  alpha_t is calculated with an analytical closed form solution.

  Args:
    beta_t: The beta parameter to use in the parameterized functions.
    rho_norm: Normalized toroidal flux coordinate.
    te_data: The temperature data to be fit, assumed to be on the face grid.

  Returns:
    The loss function for the temperature fit for a given beta_t.
  """
  alpha_t = _alpha_closed_form(
      profile_data=te_data,
      profile_edge_value=te_data[-1],
      rho_norm=rho_norm,
      beta=beta_t,
  )
  return _te_loss_fn(
      alpha_t=alpha_t,
      beta_t=beta_t,
      rho_norm=rho_norm,
      te_data=te_data,
  )


def _te_pred_fn(
    *,
    alpha_t: array_typing.ScalarFloat,
    beta_t: array_typing.ScalarFloat,
    rho_norm: array_typing.ArrayFloat,
    te_data: array_typing.ArrayFloat,
) -> array_typing.ArrayFloat:
  return (te_data[0] - te_data[-1]) * (
      (1 - rho_norm**beta_t)
  ) ** alpha_t + te_data[-1]


def _te_loss_fn(
    *,
    alpha_t: array_typing.ScalarFloat,
    beta_t: array_typing.ScalarFloat,
    rho_norm: array_typing.ArrayFloat,
    te_data: array_typing.ArrayFloat,
) -> array_typing.ScalarFloat:
  """Returns the loss function for the temperature fit.

  The fit is from the magnetic axis to rhonorm=0.9, to avoid pedestal effects.
  The pedestal will not extend to rhonorm=0.9. The choice of this value is from
  the Artaud paper.

  Args:
    alpha_t: The alpha parameter to use in the parameterized functions.
    beta_t: The beta parameter to use in the parameterized functions.
    rho_norm: Normalized toroidal flux coordinate.
    te_data: The temperature data to be fit, assumed to be on the face grid.

  Returns:
    The loss function for the temperature fit.
  """
  te_pred = _te_pred_fn(
      alpha_t=alpha_t,
      beta_t=beta_t,
      rho_norm=rho_norm,
      te_data=te_data,
  )
  mask = rho_norm < 0.9
  sliced_diff = jnp.where(mask, te_pred - te_data, 0.0)
  return jnp.sum(sliced_diff**2) / 2


def _solve_alpha_t_beta_t_grid_search(
    *,
    rho_norm: array_typing.ArrayFloat,
    te_data: array_typing.ArrayFloat,
    beta_scan_parameters: tuple[float, float, int],
) -> tuple[array_typing.ScalarFloat, array_typing.ScalarFloat]:
  """Returns the alpha and beta parameters that minimize the temperature loss function.

  Grid search is used for computational efficiency.

  Args:
    rho_norm: Normalized toroidal flux coordinate.
    te_data: The temperature data to be fit, assumed to be on the face grid.
    beta_scan_parameters: A tuple of (beta_min, beta_max, beta_grid_size)
      parameters for the grid search.

  Returns:
    The alpha and beta parameters that minimize the temperature loss function.
  """
  beta_t_trials = jnp.linspace(
      beta_scan_parameters[0],
      beta_scan_parameters[1],
      beta_scan_parameters[2],
  )

  losses = jax.vmap(_loss_for_beta_t, in_axes=(0, None, None))(
      beta_t_trials,
      rho_norm,
      te_data,
  )
  min_index = jnp.argmin(jnp.array(losses, dtype=jax_utils.get_dtype()))
  best_beta_t = beta_t_trials[min_index]
  best_alpha_t = _alpha_closed_form(
      beta=best_beta_t,
      rho_norm=rho_norm,
      profile_data=te_data,
      profile_edge_value=te_data[-1],
  )
  return best_alpha_t, best_beta_t


def cyclotron_radiation_albajar(
    static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
    dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
    geo: geometry.Geometry,
    source_name: str,
    core_profiles: state.CoreProfiles,
    unused_calculated_source_profiles: source_profiles.SourceProfiles | None,
    unused_conductivity: conductivity_base.Conductivity | None,
) -> tuple[array_typing.ArrayFloat, ...]:
  """Calculates the cyclotron radiation heat sink contribution to the electron heat equation.

  Total cyclotron radiation is from:
  F. Albajar et al 2001 Nucl. Fusion 41 665
  https://doi.org/10.1088/0029-5515/41/6/301

  Radial profile of the cyclotron radiation is from:
  J.F. Artaud et al 2018 Nucl. Fusion 58 105001
  https://doi.org/10.1088/1741-4326/aad5b1

  The alpha_n, alpha_T and beta_T parameters are calculated from best fits
  of the following electron density and temperature parameterization to the
  actual plasma data.

  n_e = n_e(0)*(1 - rhonorm**2)**alpha_n
  T_e = (T_e(0)-T_e(a))*(1 - rhonorm**beta_T)**alpha_T + T_e(a)

  Where we take a=0.9 in rhonorm space, and perform the best fit between
  0<rhonorm<0.9, to avoid pedestal effects.

  Args:
    static_runtime_params_slice: A slice of static runtime parameters.
    dynamic_runtime_params_slice: A slice of dynamic runtime parameters.
    geo: The geometry object.
    source_name: The name of the source.
    core_profiles: The core profiles object.
    unused_calculated_source_profiles: Unused.

  Returns:
    The cyclotron radiation heat sink contribution to the electron heat
    equation.
  """
  dynamic_source_runtime_params = dynamic_runtime_params_slice.sources[
      source_name
  ]
  static_source_runtime_params = static_runtime_params_slice.sources[
      source_name
  ]

  assert isinstance(dynamic_source_runtime_params, DynamicRuntimeParams)
  assert isinstance(static_source_runtime_params, StaticRuntimeParams)

  # Notation conventions based on the Albajar and Artaud papers
  # pylint: disable=invalid-name

  n_e20_face = (
      core_profiles.n_e.face_value() * core_profiles.density_reference / 1e20
  )
  n_e20_cell = core_profiles.n_e.value * core_profiles.density_reference / 1e20

  # Dimensionless optical thickness parameter, on-axis:
  # Simplified form of omega_pe**2 / (c * omega_ce) where omega_pe is the
  # plasma frequency and omega_ce is the cyclotron frequency.
  p_a_0 = 6.04e3 * geo.a_minor * n_e20_face[0] / geo.B_0

  # Dimensionless correction term for aspect ratio (equation 15 in Albajar)
  G = 0.93 * (1 + 0.85 * jnp.exp(-0.82 * geo.R_major / geo.a_minor))

  # Calculate profile fit parameters
  alpha_n = _alpha_closed_form(
      beta=2.0,
      rho_norm=static_runtime_params_slice.torax_mesh.face_centers,
      profile_data=n_e20_face,
      profile_edge_value=0.0,
  )
  beta_scan_parameters = (
      static_source_runtime_params.beta_min,
      static_source_runtime_params.beta_max,
      static_source_runtime_params.beta_grid_size,
  )
  alpha_t, beta_t = _solve_alpha_t_beta_t_grid_search(
      rho_norm=static_runtime_params_slice.torax_mesh.face_centers,
      te_data=core_profiles.T_e.face_value(),
      beta_scan_parameters=beta_scan_parameters,
  )

  # The "profile factor" (equation 13 in Albajar)
  K = (
      (alpha_n + 3.87 * alpha_t + 1.46) ** -0.79
      * (1.98 + alpha_t) ** 1.36
      * beta_t**2.14
      * (beta_t**1.53 + 1.87 * alpha_t - 0.16) ** -1.33
  )

  # Calculate power loss in [W]
  P_cycl_total = (
      3.84e-2
      * jnp.sqrt(1 - dynamic_source_runtime_params.wall_reflection_coeff)
      * geo.R_major
      * geo.a_minor**1.38
      * geo.elongation_face[-1] ** 0.79
      * geo.B_0**2.62
      * n_e20_face[0] ** 0.38
      * core_profiles.T_e.face_value()[0]
      * (16 + core_profiles.T_e.face_value()[0]) ** 2.61
      * (1 + 0.12 * core_profiles.T_e.face_value()[0] / p_a_0**0.41)
      ** -1.51
      * K
      * G
  )

  # Calculate the radial profile on the cell grid,
  # according to the Artaud formula (A.45)
  Q_cycl_shape = (
      geo.R_major
      * geo.elongation**0.79
      * (geo.F / geo.R_major) ** 2.62
      * n_e20_cell**0.38
      * core_profiles.T_e.value**3.61
  )

  # Scale the profile shape to match the total integrated power loss
  denom = math_utils.volume_integration(Q_cycl_shape, geo)
  rescaling_factor = P_cycl_total / denom
  Q_cycl = Q_cycl_shape * rescaling_factor

  return (-Q_cycl,)


@dataclasses.dataclass(kw_only=True, frozen=True, eq=True)
class CyclotronRadiationHeatSink(source.Source):
  """Cyclotron radiation heat sink for electron heat equation."""

  SOURCE_NAME: ClassVar[str] = 'cyclotron_radiation'
  model_func: source.SourceProfileFunction = cyclotron_radiation_albajar

  @property
  def source_name(self) -> str:
    return self.SOURCE_NAME

  @property
  def affected_core_profiles(self) -> tuple[source.AffectedCoreProfile, ...]:
    return (source.AffectedCoreProfile.TEMP_EL,)


class CyclotronRadiationHeatSinkConfig(base.SourceModelBase):
  """Cyclotron radiation heat sink for electron heat equation.

  Attributes:
    wall_reflection_coeff: The wall reflection coefficient is a
      machine-dependent dimensionless parameter corresponding to the fraction of
      cyclotron radiation reflected off the wall and back into the plasma where
      it is re-absorbed. The default value is a typical value.
    beta_min: The minimum value of the beta parameter used in the parameterized
      function for the temperature fit.
    beta_max: The maximum value of the beta parameter used in the parameterized
      function for the temperature fit.
    beta_grid_size: The number of points to use in the grid search for the best
      fit of the temperature function.
  """

  model_name: Literal['albajar_artaud'] = 'albajar_artaud'
  mode: runtime_params_lib.Mode = runtime_params_lib.Mode.MODEL_BASED
  wall_reflection_coeff: float = 0.9
  beta_min: float = 0.5
  beta_max: float = 8.0
  beta_grid_size: pydantic.PositiveInt = 32

  @pydantic.model_validator(mode='after')
  def _check_fields(self) -> typing_extensions.Self:
    if not self.beta_min < self.beta_max:
      raise ValueError('beta_min must be less than beta_max.')
    return self

  @property
  def model_func(self) -> source.SourceProfileFunction:
    return cyclotron_radiation_albajar

  def build_dynamic_params(
      self,
      t: chex.Numeric,
  ) -> 'DynamicRuntimeParams':
    return DynamicRuntimeParams(
        prescribed_values=tuple(
            [v.get_value(t) for v in self.prescribed_values]
        ),
        wall_reflection_coeff=self.wall_reflection_coeff,
    )

  def build_static_params(self) -> 'StaticRuntimeParams':
    return StaticRuntimeParams(
        mode=self.mode.value,
        is_explicit=self.is_explicit,
        beta_min=self.beta_min,
        beta_max=self.beta_max,
        beta_grid_size=self.beta_grid_size,
    )

  def build_source(self) -> CyclotronRadiationHeatSink:
    return CyclotronRadiationHeatSink(model_func=self.model_func)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes for representing the problem geometry."""
from collections.abc import Sequence
import dataclasses
import enum

import chex
import jax
import jax.numpy as jnp
import numpy as np
from torax._src.torax_pydantic import torax_pydantic


def face_to_cell(face: chex.Array) -> chex.Array:
  """Infers cell values corresponding to a vector of face values.

  Simply a linear interpolation between face values.

  Args:
    face: An array containing face values.

  Returns:
    cell: An array containing cell values.
  """

  return 0.5 * (face[:-1] + face[1:])


@enum.unique
class GeometryType(enum.IntEnum):
  """Integer enum for geometry type.

  This type can be used within JAX expressions to access the geometry type
  without having to call isinstance.
  """

  CIRCULAR = 0
  CHEASE = 1
  FBT = 2
  EQDSK = 3


# pylint: disable=invalid-name


@chex.dataclass(frozen=True)
class Geometry:
  r"""Describes the magnetic geometry.

  Most users should default to using the StandardGeometry class, whether the
  source of their geometry comes from CHEASE, MEQ, EQDSK, etc.

  Properties work for both 1D radial arrays and 2D stacked arrays where the
  leading dimension is time.

  Attributes:
    geometry_type: Type of geometry model used. See `GeometryType` for options.
    torax_mesh: `Grid1D` object representing the radial mesh used by TORAX.
    Phi: Toroidal magnetic flux at each radial grid point [:math:`\mathrm{Wb}`].
    Phi_face: Toroidal magnetic flux at each radial face [:math:`\mathrm{Wb}`].
    R_major: Tokamak major radius (geometric center) [:math:`\mathrm{m}`].
    a_minor: Tokamak minor radius [:math:`\mathrm{m}`].
    B_0: Vacuum toroidal magnetic field on axis [:math:`\mathrm{T}`].
    volume: Plasma volume enclosed by each flux surface on cell grid
      [:math:`\mathrm{m}^3`].
    volume_face: Plasma volume enclosed by each flux surface on face grid
      [:math:`\mathrm{m}^3`].
    area: Poloidal cross-sectional area of each flux surface on cell grid
      [:math:`\mathrm{m}^2`].
    area_face: Poloidal cross-sectional area of each flux surface on face grid
      [:math:`\mathrm{m}^2`].
    vpr: Derivative of plasma volume enclosed by each flux surface with respect
      to the normalized toroidal flux coordinate rho_norm on cell grid
      [:math:`\mathrm{m}^3`].
    vpr_face: Derivative of plasma volume enclosed by each flux surface with
      respect to the normalized toroidal flux coordinate rho_face_norm, on face
      grid [:math:`\mathrm{m}^3`].
    spr: Derivative of plasma surface area enclosed by each flux surface, with
      respect to the normalized toroidal flux coordinate rho_norm on cell grid
      [:math:`\mathrm{m}^2`]. Equal to vpr / (:math:`2 \pi` R_major).
    spr_face: Derivative of plasma surface area enclosed by each flux surface,
      with respect to the normalized toroidal flux coordinate rho_face_norm on
      face grid [:math:`\mathrm{m}^2`]. Equal to vpr_face / (:math:`2 \pi`
      R_major).
    spr_hires: Derivative of plasma surface area enclosed by each flux surface
      on a higher resolution grid, with respect to the normalized toroidal flux
      coordinate rho_norm. [:math:`\mathrm{m}^2`].
    rho_hires: Toroidal flux coordinate on a higher resolution grid
      [:math:`\mathrm{m}`].
    rho_hires_norm: Normalized toroidal flux coordinate on a higher resolution
      grid [dimensionless].
    g0: Flux surface averaged radial derivative of the plasma volume:
      :math:`\langle \nabla V \rangle` on cell grid [:math:`\mathrm{m}^2`].
    g0_face: Flux surface averaged :math:`\langle \nabla V \rangle` on the faces
      [:math:`\mathrm{m}^2`].
    g1: Flux surface averaged :math:`\langle (\nabla V)^2 \rangle` on cell grid
      [:math:`\mathrm{m}^4`].
    g1_face: Flux surface averaged :math:`\langle (\nabla V)^2 \rangle` on the
      faces [:math:`\mathrm{m}^4`].
    g2: Flux surface averaged :math:`\langle (\nabla V)^2 / R^2 \rangle` on cell
      grid [:math:`\mathrm{m}^2`], where R is the major radius along the flux
      surface being averaged.
    g2_face: Flux surface averaged :math:`\langle (\nabla V)^2 / R^2 \rangle` on
      the faces [:math:`\mathrm{m}^2`].
    g3: Flux surface averaged :math:`\langle 1 / R^2 \rangle` on cell grid
      [:math:`\mathrm{m}^{-2}`].
    g3_face: Flux surface averaged :math:`\langle 1 / R^2 \rangle` on the faces
      [:math:`\mathrm{m}^{-2}`].
    g2g3_over_rhon: Ratio of g2g3 to the normalized toroidal flux coordinate
      rho_norm on cell grid [dimensionless].
    g2g3_over_rhon_face: Ratio of g2g3 to the normalized toroidal flux
      coordinate rho_norm on face grid [dimensionless].
    g2g3_over_rhon_hires: Ratio of g2g3 to the normalized toroidal flux
      coordinate rho_norm on a higher resolution grid [dimensionless].
    F: Toroidal field flux function, :math:`F \equiv RB_\phi` on cell grid,
      where :math:`R` is major radius, and :math:`B_\phi` is the toroidal
      magnetic field [:math:`\mathrm{T m}`].
    F_face: Toroidal field flux function, :math:`F \equiv RB_\phi` on face grid
      [:math:`\mathrm{T m}`].
    F_hires: Toroidal field flux function, :math:`F \equiv RB_\phi` on the high
      resolution grid [:math:`\mathrm{T m}`].
    R_in: Radius of the flux surface at the inboard side at midplane
      [:math:`\mathrm{m}`] on cell grid. Inboard side is defined as the minimum
      radial extent of the flux surface.
    R_in_face: Radius of the flux surface at the inboard side at midplane
      [:math:`\mathrm{m}`] on face grid.
    R_out: Radius of the flux surface at the outboard side at midplane
      [:math:`\mathrm{m}`] on cell grid. Outboard side is defined as the maximum
      radial extent of the flux surface.
    R_out_face: Radius of the flux surface at the outboard side at midplane
      [:math:`\mathrm{m}`] on face grid.
    delta_face: Average of upper and lower triangularity of each flux surface at
      the faces [dimensionless]. Upper triangularity is defined as
      (R_major_local - R_upper) / a_minor_local, where R_major_local =
      (Rout+Rin)/2, a_minor_local = (Rout-Rin)/2, and R_upper is the radial
      location of the upper extent of the flux surface. Lower triangularity is
      defined as (R_major_local - R_lower) / a_minor_local, where R_lower is the
      radial location of the lower extent of the flux surface.
    elongation: Plasma elongation profile on cell grid [dimensionless].
      Elongation is defined as (Z_upper - Z_lower) / (2.0 * a_minor_local),
      where Z_upper and Z_lower are the Z coordinates of the upper and lower
      extent of the flux surface.
    elongation_face: Plasma elongation profile on face grid [dimensionless].
    Phi_b_dot: Time derivative of the toroidal magnetic flux
      [:math:`\mathrm{Wb/s}`]. Calculated across a time interval using ``Phi``
      from the Geometry objects at time t and t + dt. See
      ``torax.orchestration.step_function`` for more details.
    _z_magnetic_axis: Vertical position of the magnetic axis
      [:math:`\mathrm{m}`].
  """

  geometry_type: GeometryType
  torax_mesh: torax_pydantic.Grid1D
  Phi: chex.Array
  Phi_face: chex.Array
  R_major: chex.Array
  a_minor: chex.Array
  B_0: chex.Array
  volume: chex.Array
  volume_face: chex.Array
  area: chex.Array
  area_face: chex.Array
  vpr: chex.Array
  vpr_face: chex.Array
  spr: chex.Array
  spr_face: chex.Array
  delta_face: chex.Array
  elongation: chex.Array
  elongation_face: chex.Array
  g0: chex.Array
  g0_face: chex.Array
  g1: chex.Array
  g1_face: chex.Array
  g2: chex.Array
  g2_face: chex.Array
  g3: chex.Array
  g3_face: chex.Array
  g2g3_over_rhon: chex.Array
  g2g3_over_rhon_face: chex.Array
  g2g3_over_rhon_hires: chex.Array
  F: chex.Array
  F_face: chex.Array
  F_hires: chex.Array
  R_in: chex.Array
  R_in_face: chex.Array
  R_out: chex.Array
  R_out_face: chex.Array
  spr_hires: chex.Array
  rho_hires_norm: chex.Array
  rho_hires: chex.Array
  Phi_b_dot: chex.Array
  _z_magnetic_axis: chex.Array | None

  @property
  def q_correction_factor(self) -> chex.Numeric:
    """Ad-hoc fix for non-physical circular geometry model.

    Set such that q(r=a) = 3 for standard ITER parameters.
    """
    return jnp.where(
        self.geometry_type == GeometryType.CIRCULAR.value,
        1.25,
        1,
    )

  @property
  def rho_norm(self) -> chex.Array:
    r"""Normalized toroidal flux coordinate on cell grid [dimensionless]."""
    return self.torax_mesh.cell_centers

  @property
  def rho_face_norm(self) -> chex.Array:
    r"""Normalized toroidal flux coordinate on face grid [dimensionless]."""
    return self.torax_mesh.face_centers

  @property
  def drho_norm(self) -> chex.Array:
    r"""Grid size for rho_norm [dimensionless]."""
    return jnp.array(self.torax_mesh.dx)

  @property
  def rho_face(self) -> chex.Array:
    r"""Toroidal flux coordinate on face grid :math:`\mathrm{m}`."""
    return self.rho_face_norm * jnp.expand_dims(self.rho_b, axis=-1)

  @property
  def rho(self) -> chex.Array:
    r"""Toroidal flux coordinate on cell grid :math:`\mathrm{m}`.

    The toroidal flux coordinate is defined as
    :math:`\rho=\sqrt{\frac{\Phi}{\pi B_0}}`, where :math:`\Phi` is the
    toroidal flux enclosed by the flux surface, and :math:`B_0` the magnetic
    field on the magnetic axis.
    """
    return self.rho_norm * jnp.expand_dims(self.rho_b, axis=-1)

  @property
  def r_mid(self) -> chex.Array:
    """Midplane radius of the plasma [m], defined as (Rout-Rin)/2."""
    return (self.R_out - self.R_in) / 2

  @property
  def r_mid_face(self) -> chex.Array:
    """Midplane radius of the plasma on the face grid [m]."""
    return (self.R_out_face - self.R_in_face) / 2

  @property
  def drho(self) -> chex.Array:
    """Grid size for rho [m]."""
    return self.drho_norm * self.rho_b

  @property
  def rho_b(self) -> chex.Array:
    """Toroidal flux coordinate [m] at boundary (LCFS)."""
    return jnp.sqrt(self.Phi_b / np.pi / self.B_0)

  @property
  def Phi_b(self) -> chex.Array:
    r"""Toroidal flux at boundary (LCFS) :math:`\mathrm{Wb}`."""
    return self.Phi_face[..., -1]

  @property
  def g1_over_vpr(self) -> chex.Array:
    r"""g1/vpr [:math:`\mathrm{m}`]."""
    return self.g1 / self.vpr

  @property
  def g1_over_vpr2(self) -> chex.Array:
    r"""g1/vpr**2 [:math:`\mathrm{m}^{-2}`]."""
    return self.g1 / self.vpr**2

  @property
  def g0_over_vpr_face(self) -> jax.Array:
    """g0_face/vpr_face [:math:`m^{-1}`], equal to 1/rho_b on-axis."""
    # Calculate the bulk of the array (excluding the first element)
    # to avoid division by zero.
    bulk = self.g0_face[..., 1:] / self.vpr_face[..., 1:]
    first_element = jnp.ones_like(self.rho_b) / self.rho_b
    # Concatenate to handle both 1D (no leading dim) and 2D cases
    return jnp.concatenate(
        [jnp.expand_dims(first_element, axis=-1), bulk], axis=-1
    )

  @property
  def g1_over_vpr_face(self) -> jax.Array:
    r"""g1_face/vpr_face [:math:`\mathrm{m}`]. Zero on-axis."""
    bulk = self.g1_face[..., 1:] / self.vpr_face[..., 1:]
    first_element = jnp.zeros_like(self.rho_b)
    return jnp.concatenate(
        [jnp.expand_dims(first_element, axis=-1), bulk], axis=-1
    )

  @property
  def g1_over_vpr2_face(self) -> jax.Array:
    """g1_face/vpr_face**2 [:math:`m^{-2}`], equal to 1/rho_b**2 on-axis."""
    bulk = self.g1_face[..., 1:] / self.vpr_face[..., 1:] ** 2
    first_element = jnp.ones_like(self.rho_b) / self.rho_b**2
    return jnp.concatenate(
        [jnp.expand_dims(first_element, axis=-1), bulk], axis=-1
    )

  def z_magnetic_axis(self) -> chex.Numeric:
    """z position of magnetic axis [m]."""
    z_magnetic_axis = self._z_magnetic_axis
    if z_magnetic_axis is not None:
      return z_magnetic_axis
    else:
      raise ValueError('Geometry does not have a z magnetic axis.')


def stack_geometries(geometries: Sequence[Geometry]) -> Geometry:
  """Batch together a sequence of geometries.

  Args:
    geometries: A sequence of geometries to stack. The geometries must have the
      same mesh, geometry type.

  Returns:
    A Geometry object, where each array attribute has an additional
    leading axis (e.g. for the time dimension) compared to each Geometry in
    the input sequence.
  """
  if not geometries:
    raise ValueError('No geometries provided.')
  # Ensure that all geometries have same mesh and are of same type.
  first_geo = geometries[0]
  torax_mesh = first_geo.torax_mesh
  geometry_type = first_geo.geometry_type
  for geometry in geometries[1:]:
    if geometry.torax_mesh != torax_mesh:
      raise ValueError('All geometries must have the same mesh.')
    if geometry.geometry_type != geometry_type:
      raise ValueError('All geometries must have the same geometry type.')

  stacked_data = {}
  for field in dataclasses.fields(first_geo):
    field_name = field.name
    field_value = getattr(first_geo, field_name)
    # Stack stackable fields. Save first geo's value for non-stackable fields.
    if isinstance(field_value, chex.Array):
      field_values = [getattr(geo, field_name) for geo in geometries]
      stacked_data[field_name] = jnp.stack(field_values)
    else:
      stacked_data[field_name] = field_value
  # Create a new object with the stacked data with the same class (i.e.
  # could be child classes of Geometry)
  return first_geo.__class__(**stacked_data)


def update_geometries_with_Phibdot(
    *,
    dt: chex.Numeric,
    geo_t: Geometry,
    geo_t_plus_dt: Geometry,
) -> tuple[Geometry, Geometry]:
  """Update Phibdot in the geometry dataclasses used in the time interval.

  Phibdot is used in calc_coeffs to calculate terms related to time-dependent
  geometry. It should be set to be the same for geo_t and geo_t_plus_dt for
  each given time interval. This means that geo_t_plus_dt.Phibdot will not
  necessarily be the same as the geo_t.Phibdot at the next time step.

  Args:
    dt: Time step duration.
    geo_t: The geometry of the torus during this time step of the simulation.
    geo_t_plus_dt: The geometry of the torus during the next time step of the
      simulation.

  Returns:
    Tuple containing:
      - The geometry of the torus during this time step of the simulation.
      - The geometry of the torus during the next time step of the simulation.
  """
  Phibdot = (geo_t_plus_dt.Phi_b - geo_t.Phi_b) / dt
  geo_t = dataclasses.replace(geo_t, Phi_b_dot=Phibdot)
  geo_t_plus_dt = dataclasses.replace(geo_t_plus_dt, Phi_b_dot=Phibdot)
  return geo_t, geo_t_plus_dt



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module for a single source/sink term.

This module contains all the base classes for defining source terms. Other files
in this folder use these classes to define specific types of sources/sinks.

See Source class docstring for more details on what a TORAX source is and how to
use it.
"""
import abc
import dataclasses
import enum
import typing
from typing import ClassVar, Protocol

import chex
from jax import numpy as jnp
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.sources import runtime_params as runtime_params_lib
from torax._src.sources import source_profiles


@typing.runtime_checkable
class SourceProfileFunction(Protocol):
  """Sources implement these functions to be able to provide source profiles."""

  def __call__(
      self,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo: geometry.Geometry,
      source_name: str,
      core_profiles: state.CoreProfiles,
      calculated_source_profiles: source_profiles.SourceProfiles | None,
      unused_conductivity: conductivity_base.Conductivity | None,
  ) -> tuple[chex.Array, ...]:
    ...


@enum.unique
class AffectedCoreProfile(enum.IntEnum):
  """Defines which part of the core profiles the source helps evolve.

  The profiles of each source/sink are terms included in equations evolving
  different core profiles. This enum maps a source to those equations.
  """

  # Current density equation.
  PSI = 1
  # Electron density equation.
  NE = 2
  # Ion temperature equation.
  TEMP_ION = 3
  # Electron temperature equation.
  TEMP_EL = 4


@dataclasses.dataclass(kw_only=True, frozen=True, eq=True)
class Source(abc.ABC):
  """Base class for a single source/sink term.

  Sources are used to compute source profiles (see source_profiles.py), which
  are in turn used to compute coeffs in sim.py.

  Attributes:
    SOURCE_NAME: The name of the source.
    runtime_params: Input dataclass containing all the source-specific runtime
      parameters. At runtime, the parameters here are interpolated to a specific
      time t and then passed to the model_func, depending on the mode this
      source is running in.
    affected_core_profiles: Core profiles affected by this source's profile(s).
      This attribute defines which equations the source profiles are terms for.
      By default, the number of affected core profiles should equal the rank of
      the output shape returned by `output_shape`.
    model_func: The function used when the runtime type is set to
      "MODEL_BASED". If not provided, then it defaults to returning zeros.
    affected_core_profiles_ints: Derived property from the
      affected_core_profiles. Integer values of those enums.
  """

  SOURCE_NAME: ClassVar[str] = 'source'
  model_func: SourceProfileFunction | None = None

  @property
  @abc.abstractmethod
  def source_name(self) -> str:
    """Returns the name of the source."""

  @property
  @abc.abstractmethod
  def affected_core_profiles(self) -> tuple[AffectedCoreProfile, ...]:
    """Returns the core profiles affected by this source."""

  def get_value(
      self,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
      calculated_source_profiles: source_profiles.SourceProfiles | None,
      conductivity: conductivity_base.Conductivity | None,
  ) -> tuple[chex.Array, ...]:
    """Returns the cell grid profile for this source during one time step.

    Args:
      static_runtime_params_slice: Static runtime parameters.
      dynamic_runtime_params_slice: Slice of the general TORAX config that can
        be used as input for this time step.
      geo: Geometry of the torus.
      core_profiles: Core plasma profiles. May be the profiles at the start of
        the time step or a "live" set of core profiles being actively updated
        depending on whether this source is explicit or implicit. Explicit
        sources get the core profiles at the start of the time step, implicit
        sources get the "live" profiles that is updated through the course of
        the time step as the solver converges.
      calculated_source_profiles: The source profiles which have already been
        calculated for this time step if they exist. This is used to avoid
        recalculating profiles that are used as inputs to other sources. These
        profiles will only exist for Source instances that are implicit. i.e.
        explicit sources cannot depend on other calculated source profiles. In
        addition, different source types will have different availability of
        specific calculated_source_profiles since the calculation order matters.
        See source_profile_builders.py for more details.
      conductivity: Conductivity profile if it exists. It is only provided for
        implicit sources.

    Returns:
      A tuple of arrays of shape (cell grid length,) with one array per affected
      core profile.
    """
    dynamic_source_runtime_params = dynamic_runtime_params_slice.sources[
        self.source_name
    ]

    mode = static_runtime_params_slice.sources[self.source_name].mode
    match mode:
      case runtime_params_lib.Mode.MODEL_BASED.value:
        if self.model_func is None:
          raise ValueError(
              'Source is in MODEL_BASED mode but has no model function.'
          )
        return self.model_func(
            static_runtime_params_slice,
            dynamic_runtime_params_slice,
            geo,
            self.source_name,
            core_profiles,
            calculated_source_profiles,
            conductivity,
        )
      case runtime_params_lib.Mode.PRESCRIBED.value:
        if len(self.affected_core_profiles) != len(
            dynamic_source_runtime_params.prescribed_values
        ):
          raise ValueError(
              'When using PRESCRIBED mode, the number of prescribed values must'
              ' match the number of affected core profiles. Was: '
              f'{len(dynamic_source_runtime_params.prescribed_values)} '
              f' Expected: {len(self.affected_core_profiles)}.'
          )
        return dynamic_source_runtime_params.prescribed_values
      case runtime_params_lib.Mode.ZERO.value:
        zeros = jnp.zeros(geo.rho_norm.shape)
        return (zeros,) * len(self.affected_core_profiles)
      case _:
        raise ValueError(f'Unknown mode: {mode}')

  def __hash__(self) -> int:
    return hash((self.SOURCE_NAME, self.model_func))

  def __eq__(self, other) -> bool:
    return (
        isinstance(other, type(self))
        and self.SOURCE_NAME == other.SOURCE_NAME
        and self.model_func == other.model_func
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The ChiTimeStepCalculator class.

Steps through time using a heuristic based on chi_max.
"""

import functools

import jax
from jax import numpy as jnp
from torax._src import jax_utils
from torax._src import state as state_module
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.time_step_calculator import time_step_calculator


class ChiTimeStepCalculator(time_step_calculator.TimeStepCalculator):
  """TimeStepCalculator based on chi_max heuristic.

  Attributes:
    config: General configuration parameters.
  """

  @functools.partial(jax_utils.jit, static_argnames=['self'])
  def next_dt(
      self,
      dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo: geometry.Geometry,
      core_profiles: state_module.CoreProfiles,
      core_transport: state_module.CoreTransport,
  ) -> jax.Array:
    """Calculates the next time step duration.

    This calculation is a heuristic scaling of the maximum stable step
    size for the explicit method, and is therefore a function of chi_max.

    Args:
      dynamic_runtime_params_slice: Input runtime parameters that can change
        without triggering a JAX recompilation.
      geo: Geometry for the tokamak being simulated.
      core_profiles: Current core plasma profiles.
      core_transport: Used to calculate maximum step size.

    Returns:
      dt: Scalar time step duration.
    """

    chi_max = core_transport.chi_max(geo)

    basic_dt = (3.0 / 4.0) * (geo.drho_norm**2) / chi_max

    dt = jnp.minimum(
        dynamic_runtime_params_slice.numerics.chi_timestep_prefactor * basic_dt,
        dynamic_runtime_params_slice.numerics.max_dt,
    )

    return dt



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Sauter model for bootstrap current."""

from typing import Literal

import chex
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry as geometry_lib
from torax._src.neoclassical.bootstrap_current import base
from torax._src.neoclassical.bootstrap_current import runtime_params
from torax._src.sources import bootstrap_current_source


@chex.dataclass(frozen=True)
class DynamicRuntimeParams(runtime_params.DynamicRuntimeParams):
  """Dynamic runtime params for the Sauter model."""

  bootstrap_multiplier: float


class SauterModel(base.BootstrapCurrentModel):
  """Sauter model for bootstrap current."""

  def calculate_bootstrap_current(
      self,
      dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
      geometry: geometry_lib.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> base.BootstrapCurrent:
    """Calculates bootstrap current."""
    bootstrap_params = (
        dynamic_runtime_params_slice.neoclassical.bootstrap_current
    )
    assert isinstance(bootstrap_params, DynamicRuntimeParams)
    result = bootstrap_current_source.calc_sauter_model(
        bootstrap_multiplier=bootstrap_params.bootstrap_multiplier,
        density_reference=dynamic_runtime_params_slice.numerics.density_reference,
        Z_eff_face=dynamic_runtime_params_slice.plasma_composition.Z_eff_face,
        Z_i_face=core_profiles.Z_i_face,
        n_e=core_profiles.n_e,
        n_i=core_profiles.n_i,
        T_e=core_profiles.T_e,
        T_i=core_profiles.T_i,
        psi=core_profiles.psi,
        q_face=core_profiles.q_face,
        geo=geometry,
    )
    return base.BootstrapCurrent(
        j_bootstrap=result.j_bootstrap,
        j_bootstrap_face=result.j_bootstrap_face,
    )

  def __eq__(self, other) -> bool:
    return isinstance(other, self.__class__)

  def __hash__(self) -> int:
    return hash(self.__class__)


class SauterModelConfig(base.BootstrapCurrentModelConfig):
  """Config for the Sauter model implementation of bootstrap current.

  Attributes:
    bootstrap_multiplier: Multiplication factor for bootstrap current.
  """

  model_name: Literal['sauter'] = 'sauter'
  bootstrap_multiplier: float = 1.0

  def build_dynamic_params(self) -> DynamicRuntimeParams:
    return DynamicRuntimeParams(bootstrap_multiplier=self.bootstrap_multiplier)

  def build_model(self) -> SauterModel:
    return SauterModel()



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Physics formulas mostly related to CoreProfiles calculations.

Functions:
    - calculate_main_ion_dilution_factor: Calculates the main ion dilution
      factor based on average impurity charge and Z_eff.
    - calculate_pressure: Calculates pressure from density and temperatures.
    - calc_pprime: Calculates total pressure gradient with respect to poloidal
      flux.
    - calc_FFprime: Calculates FF', an output quantity used for equilibrium
      coupling.
    - calculate_stored_thermal_energy: Calculates stored thermal energy from
      pressures.
    - calculate_greenwald_fraction: Calculates the Greenwald fraction from the
      averaged electron density (can be line-averaged or volume-averaged).
"""
import jax
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import constants
from torax._src import state
from torax._src.geometry import geometry

_trapz = jax.scipy.integrate.trapezoid

# pylint: disable=invalid-name


# TODO(b/377225415): generalize to arbitrary number of ions.
def calculate_main_ion_dilution_factor(
    Z_i: array_typing.ScalarFloat,
    Z_impurity: array_typing.ArrayFloat,
    Z_eff: array_typing.ArrayFloat,
) -> jax.Array:
  """Calculates the main ion dilution factor based on a single assumed impurity and general main ion charge."""
  return (Z_impurity - Z_eff) / (Z_i * (Z_impurity - Z_i))


def calculate_pressure(
    core_profiles: state.CoreProfiles,
) -> tuple[array_typing.ArrayFloat, ...]:
  """Calculates pressure from density and temperatures on the face grid.

  Args:
    core_profiles: CoreProfiles object containing information on temperatures
      and densities.

  Returns:
    pressure_thermal_el_face: Electron thermal pressure [Pa]
    pressure_thermal_ion_face: Ion thermal pressure [Pa]
    pressure_thermal_tot_face: Total thermal pressure [Pa]
  """
  n_e = core_profiles.n_e.face_value()
  n_i = core_profiles.n_i.face_value()
  n_impurity = core_profiles.n_impurity.face_value()
  T_i = core_profiles.T_i.face_value()
  T_e = core_profiles.T_e.face_value()
  prefactor = constants.CONSTANTS.keV2J * core_profiles.density_reference
  pressure_thermal_el_face = n_e * T_e * prefactor
  pressure_thermal_ion_face = (n_i + n_impurity) * T_i * prefactor
  pressure_thermal_tot_face = (
      pressure_thermal_el_face + pressure_thermal_ion_face
  )
  return (
      pressure_thermal_el_face,
      pressure_thermal_ion_face,
      pressure_thermal_tot_face,
  )


def calc_pprime(
    core_profiles: state.CoreProfiles,
) -> array_typing.ArrayFloat:
  r"""Calculates total pressure gradient with respect to poloidal flux.

  Args:
    core_profiles: CoreProfiles object containing information on temperatures
      and densities.

  Returns:
    pprime: Total pressure gradient :math:`\partial p / \partial \psi`
      with respect to the normalized toroidal flux coordinate, on the face grid.
  """

  prefactor = constants.CONSTANTS.keV2J * core_profiles.density_reference

  _, _, p_total = calculate_pressure(core_profiles)
  psi = core_profiles.psi.face_value()
  n_e = core_profiles.n_e.face_value()
  n_i = core_profiles.n_i.face_value()
  n_impurity = core_profiles.n_impurity.face_value()
  T_i = core_profiles.T_i.face_value()
  T_e = core_profiles.T_e.face_value()
  dne_drhon = core_profiles.n_e.face_grad()
  dni_drhon = core_profiles.n_i.face_grad()
  dnimp_drhon = core_profiles.n_impurity.face_grad()
  dti_drhon = core_profiles.T_i.face_grad()
  dte_drhon = core_profiles.T_e.face_grad()
  dpsi_drhon = core_profiles.psi.face_grad()

  dptot_drhon = prefactor * (
      n_e * dte_drhon
      + n_i * dti_drhon
      + n_impurity * dti_drhon
      + dne_drhon * T_e
      + dni_drhon * T_i
      + dnimp_drhon * T_i
  )

  # Calculate on-axis value with L'Hôpital's rule using 2nd order forward
  # difference approximation for second derivative at edge.
  pprime_face_axis = jnp.expand_dims(
      (2 * p_total[0] - 5 * p_total[1] + 4 * p_total[2] - p_total[3])
      / (2 * psi[0] - 5 * psi[1] + 4 * psi[2] - psi[3]),
      axis=0,
  )

  # Zero on-axis due to boundary conditions. Avoid division by zero.
  pprime_face = jnp.concatenate(
      [pprime_face_axis, dptot_drhon[1:] / dpsi_drhon[1:]]
  )

  return pprime_face


def calc_FFprime(
    core_profiles: state.CoreProfiles,
    geo: geometry.Geometry,
) -> array_typing.ArrayFloat:
  r"""Calculates FF', an output quantity used for equilibrium coupling.

  Calculation is based on the following formulation of the magnetic
  equilibrium equation:
  :math:`-j_{tor} = 2\pi (Rp' + \frac{1}{\mu_0 R}FF')`

  And following division by R and flux surface averaging:

  :math:`-\langle \frac{j_{tor}}{R} \rangle = 2\pi (p' +
  \langle\frac{1}{R^2}\rangle\frac{FF'}{\mu_0})`

  Args:
    core_profiles: CoreProfiles object containing information on temperatures
      and densities.
    geo: Magnetic equilibrium.

  Returns:
    FFprime:   F is the toroidal flux function, and F' is its derivative with
      respect to the poloidal flux.
  """

  mu0 = constants.CONSTANTS.mu0
  pprime = calc_pprime(core_profiles)
  # g3 = <1/R^2>
  g3 = geo.g3_face
  jtor_over_R = core_profiles.j_total_face / geo.R_major

  FFprime_face = -(jtor_over_R / (2 * jnp.pi) + pprime) * mu0 / g3
  return FFprime_face


def calculate_stored_thermal_energy(
    p_el: array_typing.ArrayFloat,
    p_ion: array_typing.ArrayFloat,
    p_tot: array_typing.ArrayFloat,
    geo: geometry.Geometry,
) -> tuple[array_typing.ScalarFloat, ...]:
  """Calculates stored thermal energy from pressures.

  Args:
    p_el: Electron pressure [Pa]
    p_ion: Ion pressure [Pa]
    p_tot: Total pressure [Pa]
    geo: Geometry object

  Returns:
    wth_el: Electron thermal stored energy [J]
    wth_ion: Ion thermal stored energy [J]
    wth_tot: Total thermal stored energy [J]
  """
  wth_el = _trapz(1.5 * p_el * geo.vpr_face, geo.rho_face_norm)
  wth_ion = _trapz(1.5 * p_ion * geo.vpr_face, geo.rho_face_norm)
  wth_tot = _trapz(1.5 * p_tot * geo.vpr_face, geo.rho_face_norm)

  return wth_el, wth_ion, wth_tot


def calculate_greenwald_fraction(
    n_e_avg: array_typing.ScalarFloat,
    core_profiles: state.CoreProfiles,
    geo: geometry.Geometry,
) -> array_typing.ScalarFloat:
  """Calculates the Greenwald fraction from the averaged electron density.

  Different averaging can be used, e.g. volume-averaged or line-averaged.

  Args:
    n_e_avg: Averaged electron density [density_reference m^-3]
    core_profiles: CoreProfiles object containing information on currents and
      densities.
    geo: Geometry object

  Returns:
    fgw: Greenwald density fraction
  """
  # gw_limit is in units of 10^20 m^-3 when Ip is in MA and a_minor is in m.
  gw_limit = (
      core_profiles.Ip_profile_face[-1] * 1e-6 / (jnp.pi * geo.a_minor**2)
  )
  fgw = n_e_avg * core_profiles.density_reference / (gw_limit * 1e20)
  return fgw



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A wrapper for QLKNN transport surrogate models."""
from collections.abc import Mapping
from typing import Final

from fusion_surrogates.qlknn import qlknn_model
import immutabledict
import jax
import jax.numpy as jnp
from torax._src import jax_utils
from torax._src.transport_model import base_qlknn_model
from torax._src.transport_model import qualikiz_based_transport_model


# Convert flux names from Qualikiz to TORAX.
_FLUX_NAME_MAP: Final[Mapping[str, str]] = immutabledict.immutabledict({
    'efiITG': 'qi_itg',
    'efeITG': 'qe_itg',
    'pfeITG': 'pfe_itg',
    'efeTEM': 'qe_tem',
    'efiTEM': 'qi_tem',
    'pfeTEM': 'pfe_tem',
    'efeETG': 'qe_etg',
})


class QLKNNModelWrapper(base_qlknn_model.BaseQLKNNModel):
  """A TORAX wrapper for a QLKNN Model from the fusion_surrogates library."""

  def __init__(
      self,
      path: str,
      name: str = '',
      flux_name_map: Mapping[str, str] | None = None,
  ):
    if flux_name_map is None:
      flux_name_map = _FLUX_NAME_MAP
    self._flux_name_map = flux_name_map

    if path:
      self._model = qlknn_model.QLKNNModel.load_model_from_path(path, name)
    elif name:
      self._model = qlknn_model.QLKNNModel.load_model_from_name(name)
    else:
      self._model = qlknn_model.QLKNNModel.load_default_model()
    super().__init__(path=self._model.path, name=self._model.name)

  @property
  def inputs_and_ranges(self) -> base_qlknn_model.InputsAndRanges:
    return self._model.inputs_and_ranges

  def get_model_inputs_from_qualikiz_inputs(
      self, qualikiz_inputs: qualikiz_based_transport_model.QualikizInputs
  ) -> jax.Array:
    """Converts QualikizInputs to model inputs."""
    # Non-trivial mappings from QualikizInputs to model inputs.
    input_map = {
        'Ani': lambda x: x.Ani0,
        'LogNuStar': lambda x: x.log_nu_star_face,
    }

    def _get_input(key: str) -> jax.Array:
      # If no complex mapping is defined, we use the trivial mapping.
      return jnp.array(
          input_map.get(key, lambda x: getattr(x, key))(qualikiz_inputs),
          dtype=jax_utils.get_dtype(),
      )

    return jnp.array(
        [_get_input(key) for key in self.inputs_and_ranges.keys()],
        dtype=jax_utils.get_dtype(),
    ).T

  def predict(self, inputs: jax.Array) -> dict[str, jax.Array]:
    """Predicts the fluxes given the inputs."""
    model_predictions = self._model.predict(inputs)

    return {
        self._flux_name_map.get(flux_name, flux_name): flux_value
        for flux_name, flux_value in model_predictions.items()
    }



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default plotting configuration for Torax runs."""

from torax._src.plotting import plotruns_lib

PLOT_CONFIG = plotruns_lib.FigureProperties(
    rows=4,
    cols=4,
    tick_fontsize=8,
    axes_fontsize=8,
    default_legend_fontsize=7,
    figure_size_factor=5,
    title_fontsize=12,
    axes=(
        # For chi, set histogram percentile for y-axis upper limit, due to
        # volatile nature of the data. Do not include first timepoint since
        # chi is defined as zero there and may unduly affect ylim.
        plotruns_lib.PlotProperties(
            attrs=('T_i', 'T_e'),
            labels=(r'$T_\mathrm{i}$', r'$T_\mathrm{e}$'),
            ylabel='Temperature [keV]',
        ),
        plotruns_lib.PlotProperties(
            attrs=('n_e', 'n_i'),
            labels=(r'$n_\mathrm{e}$', r'$n_\mathrm{i}$'),
            ylabel=r'Density $[10^{20}~m^{-3}]$',
        ),
        plotruns_lib.PlotProperties(
            attrs=('chi_turb_i', 'chi_turb_e'),
            labels=(r'$\chi_\mathrm{i}$', r'$\chi_\mathrm{e}$'),
            ylabel=r'Heat conductivity $[m^2/s]$',
            upper_percentile=98.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
        ),
        plotruns_lib.PlotProperties(
            attrs=('D_turb_e', 'V_turb_e'),
            labels=(r'$D_\mathrm{e}$', r'$V_\mathrm{e}$'),
            ylabel=r'Diff $[m^2/s]$ or Conv $[m/s]$',
            upper_percentile=98.0,
            lower_percentile=2.0,
            include_first_timepoint=False,
            ylim_min_zero=False,
        ),
        plotruns_lib.PlotProperties(
            plot_type=plotruns_lib.PlotType.TIME_SERIES,
            attrs=('Ip_profile', 'I_bootstrap', 'I_aux_generic', 'I_ecrh'),
            labels=(
                r'$I_\mathrm{p}$',
                r'$I_\mathrm{bs}$',
                r'$I_\mathrm{generic}$',
                r'$I_\mathrm{ecrh}$',
            ),
            ylabel=r'Current [MA]',
            suppress_zero_values=True,  # Do not plot all-zero data
        ),
        plotruns_lib.PlotProperties(
            attrs=('psi',),
            labels=(r'$\psi$',),
            ylabel=r'Poloidal flux [Wb]',
        ),
        plotruns_lib.PlotProperties(
            attrs=(
                'j_total',
                'j_ohmic',
                'j_bootstrap',
                'j_generic_current',
                'j_ecrh',
            ),
            labels=(
                r'$j_\mathrm{tot}$',
                r'$j_\mathrm{ohm}$',
                r'$j_\mathrm{bs}$',
                r'$j_\mathrm{generic}$',
                r'$j_\mathrm{ecrh}$',
            ),
            ylabel=r'Toroidal current $[MA~m^{-2}]$',
            legend_fontsize=7,  # Smaller fontsize for this plot
            suppress_zero_values=True,  # Do not plot all-zero data
        ),
        plotruns_lib.PlotProperties(
            attrs=('q',),
            labels=(r'$q$',),
            ylabel='Safety factor',
        ),
        plotruns_lib.PlotProperties(
            attrs=('magnetic_shear',),
            labels=(r'$\hat{s}$',),
            ylabel='Magnetic shear',
        ),
        plotruns_lib.PlotProperties(
            plot_type=plotruns_lib.PlotType.TIME_SERIES,
            attrs=('Q_fusion',),
            labels=(r'$Q_\mathrm{fusion}$',),
            ylabel='Fusion gain',
        ),
        plotruns_lib.PlotProperties(
            attrs=('v_loop',),
            labels=(r'$\dot{\psi}$',),
            ylabel='Loop voltage',
            upper_percentile=98.0,
        ),
        plotruns_lib.PlotProperties(
            attrs=(
                'p_icrh_i',
                'p_icrh_e',
                'p_ecrh_e',
                'p_generic_heat_i',
                'p_generic_heat_e',
            ),
            labels=(
                r'$Q_\mathrm{ICRH,i}$',
                r'$Q_\mathrm{ICRH,e}$',
                r'$Q_\mathrm{ECRH,e}$',
                r'$Q_\mathrm{generic,i}$',
                r'$Q_\mathrm{generic,e}$',
            ),
            ylabel=r'External heat source density $[MW~m^{-3}]$',
            legend_fontsize=7,  # Smaller fontsize for this plot
            suppress_zero_values=True,  # Do not plot all-zero data
        ),
        plotruns_lib.PlotProperties(
            attrs=('p_alpha_i', 'p_alpha_e', 'p_ohmic_e', 'ei_exchange'),
            labels=(
                r'$Q_\mathrm{alpha,i}$',
                r'$Q_\mathrm{alpha,e}$',
                r'$Q_\mathrm{ohmic}$',
                r'$Q_\mathrm{ei}$',
            ),
            ylabel=r'Internal heat source density $[MW~m^{-3}]$',
            legend_fontsize=6,  # Smaller fontsize for this plot
            suppress_zero_values=True,  # Do not plot all-zero data
        ),
        plotruns_lib.PlotProperties(
            attrs=(
                'p_bremsstrahlung_e',
                'p_impurity_radiation_e',
                'p_cyclotron_radiation_e',
            ),
            labels=(
                r'$Q_\mathrm{brems}$',
                r'$Q_\mathrm{rad}$',
                r'$Q_\mathrm{cycl}$',
            ),
            ylabel=r'Heat sink density $[MW~m^{-3}]$',
            suppress_zero_values=True,  # Do not plot all-zero data
        ),
        plotruns_lib.PlotProperties(
            plot_type=plotruns_lib.PlotType.TIME_SERIES,
            attrs=(
                'P_auxiliary',
                'P_ohmic_e',
                'P_alpha_total',
                'P_bremsstrahlung_e',
                'P_radiation_e',
                'P_cyclotron_e',
            ),
            labels=(
                r'$P_\mathrm{aux}$',
                r'$P_\mathrm{ohm}$',
                r'$P_\mathrm{\alpha}$',
                r'$P_\mathrm{brems}$',
                r'$P_\mathrm{rad}$',
                r'$P_\mathrm{cycl}$',
            ),
            ylabel=r'Total heating/sink powers $[MW]$',
            legend_fontsize=6,  # Smaller fontsize for this plot
            suppress_zero_values=True,  # Do not plot all-zero data
        ),
        plotruns_lib.PlotProperties(
            attrs=('Z_impurity',),
            labels=(r'$\langle Z_{impurity} \rangle$',),
            ylabel=r'Average impurity charge [amu]',
        ),
    ),
)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Routines for calculating impurity radiation based on a polynomial fit."""

import functools
from typing import Final, Literal, Mapping, Sequence

import chex
import immutabledict
import jax.numpy as jnp
import numpy as np
from torax._src import array_typing
from torax._src import constants
from torax._src import jax_utils
from torax._src import state
from torax._src.config import plasma_composition
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.sources import base
from torax._src.sources import runtime_params as runtime_params_lib
from torax._src.sources import source as source_lib
from torax._src.sources import source_profiles
from torax._src.sources.impurity_radiation_heat_sink import impurity_radiation_heat_sink


# Default value for the model function to be used for the impurity radiation
# source. This is also used as an identifier for the model function in
# the source config for Pydantic to "discriminate" against.
DEFAULT_MODEL_FUNCTION_NAME: str = 'mavrin_fit'

# Polynomial fit coefficients from A. A. Mavrin (2018):
# Improved fits of coronal radiative cooling rates for high-temperature plasmas,
# Radiation Effects and Defects in Solids, 173:5-6, 388-398,
# DOI: 10.1080/10420150.2018.1462361
_MAVRIN_L_COEFFS: Final[Mapping[str, array_typing.ArrayFloat]] = (
    immutabledict.immutabledict({
        'He3': np.array([
            [2.5020e-02, -9.3730e-02, 1.0156e-01, 3.1469e-01, -3.5551e01],
        ]),
        'He4': np.array([
            [2.5020e-02, -9.3730e-02, 1.0156e-01, 3.1469e-01, -3.5551e01],
        ]),
        'Li': np.array([
            [3.5190e-02, -1.6070e-01, 2.5082e-01, 1.9475e-01, -3.5115e01],
        ]),
        'Be': np.array([
            [4.1690e-02, -2.1384e-01, 3.8363e-01, 3.7270e-02, -3.4765e01],
        ]),
        'C': np.array([
            [-7.2904e00, -1.6637e01, -1.2788e01, -5.0085e00, -3.4738e01],
            [4.4470e-02, -2.9191e-01, 6.8856e-01, -3.6687e-01, -3.4174e01],
        ]),
        'N': np.array([
            [-6.9621e00, -1.1570e01, -6.0605e00, -2.3614e00, -3.4065e01],
            [5.8770e-02, -1.7160e-01, 7.6272e-01, -5.9668e-01, -3.3899e01],
            [2.8350e-02, -2.2790e-01, 7.0047e-01, -5.2628e-01, -3.3913e01],
        ]),
        'O': np.array([
            [0.0000e00, -5.3765e00, -1.7141e01, -1.5635e01, -3.7257e01],
            [1.4360e-02, -2.0850e-01, 7.9655e-01, -7.6211e-01, -3.3640e01],
        ]),
        'Ne': np.array([
            [1.5648e01, 2.8939e01, 1.5230e01, 1.7309e00, -3.3132e01],
            [1.7244e-01, -3.9544e-01, 8.6842e-01, -8.7750e-01, -3.3290e01],
            [-2.6930e-02, 4.3960e-02, 2.9731e-01, -4.5345e-01, -3.3410e01],
        ]),
        'Ar': np.array([
            [1.5353e01, 3.9161e01, 3.0769e01, 6.5221e00, -3.2155e01],
            [4.9806e00, -7.6887e00, 1.5389e00, 5.4490e-01, -3.2530e01],
            [-8.2260e-02, 1.7480e-01, 6.1339e-01, -1.6674e00, -3.1853e01],
        ]),
        'Kr': np.array([
            [-1.3564e01, -4.0133e01, -4.4723e01, -2.1484e01, -3.4512e01],
            [-5.2704e00, -2.5865e00, 1.9148e00, -5.0091e-01, -3.1399e01],
            [4.8356e-01, -2.9674e00, 6.6831e00, -6.3683e00, -2.9954e01],
        ]),
        'Xe': np.array([
            [2.5615e01, 5.9580e01, 4.7081e01, 1.4351e01, -2.9303e01],
            [1.0748e01, -1.1628e01, 1.2808e00, 5.9339e-01, -3.1113e01],
            [1.0069e01, -3.6885e01, 4.8614e01, -2.7526e01, -2.5813e01],
            [1.0858e00, -7.5181e00, 1.9619e01, -2.2592e01, -2.2138e01],
        ]),
        'W': np.array([
            [-1.0103e-01, -1.0311e00, -9.5126e-01, 3.8304e-01, -3.0374e01],
            [5.1849e01, -6.3303e01, 2.2824e01, -2.9208e00, -3.0238e01],
            [-3.6759e-01, 2.6627e00, -6.2740e00, 5.2499e00, -3.2153e01],
        ]),
    })
)

# Temperature boundaries in keV, separating the rows for the fit coefficients.
_TEMPERATURE_INTERVALS: Final[Mapping[str, array_typing.ArrayFloat]] = (
    immutabledict.immutabledict({
        'C': np.array([0.5]),
        'N': np.array([0.5, 2.0]),
        'O': np.array([0.3]),
        'Ne': np.array([0.7, 5.0]),
        'Ar': np.array([0.6, 3.0]),
        'Kr': np.array([0.447, 2.364]),
        'Xe': np.array([0.5, 2.5, 10.0]),
        'W': np.array([1.5, 4.0]),
    })
)


# pylint: disable=invalid-name
def _calculate_impurity_radiation_single_species(
    T_e: array_typing.ArrayFloat,
    ion_symbol: str,
) -> array_typing.ArrayFloat:
  """Calculates the line radiation for single impurity species.

  Polynomial fit range is 0.1-100 keV, which is well within the typical
  bounds of core tokamak modelling. For safety, inputs are clipped to avoid
  extrapolation outside this range.

  Args:
    T_e: Electron temperature [keV].
    ion_symbol: Species to calculate line radiation for.

  Returns:
    LZ: Radiative cooling rate in units of Wm^3.
  """

  if ion_symbol not in constants.ION_SYMBOLS:
    raise ValueError(
        f'Invalid ion symbol: {ion_symbol}. Allowed symbols are :'
        f' {constants.ION_SYMBOLS}'
    )

  # Avoid extrapolating fitted polynomial out of bounds.
  T_e = jnp.clip(T_e, 0.1, 100.0)

  # Gather coefficients for each temperature
  if ion_symbol in {'He3', 'He4', 'Be', 'Li'}:
    interval_indices = 0
  else:
    interval_indices = jnp.searchsorted(_TEMPERATURE_INTERVALS[ion_symbol], T_e)

  L_coeffs_in_range = jnp.take(
      _MAVRIN_L_COEFFS[ion_symbol], interval_indices, axis=0
  ).transpose()

  X = jnp.log10(T_e)
  log10_LZ = jnp.polyval(L_coeffs_in_range, X)
  return 10**log10_LZ


@functools.partial(
    jax_utils.jit,
    static_argnames=[
        'ion_symbols',
    ],
)
def calculate_total_impurity_radiation(
    ion_symbols: Sequence[str],
    ion_mixture: plasma_composition.DynamicIonMixture,
    T_e: array_typing.ArrayFloat,
) -> array_typing.ArrayFloat:
  """Calculates impurity line radiation profile (JAX-compatible).

  Args:
    ion_symbols: Ion symbols of the impurity species.
    ion_mixture: DynamicIonMixture object containing impurity information.
    T_e: Electron temperature [keV]. Can be any sized array, e.g. on cell grid,
      face grid, or a single scalar.

  Returns:
    effective_LZ: Total effective radiative cooling rate in units of Wm^3,
      summed over all species in the mixture. The shape of LZ is the same as
      T_e.
  """

  effective_LZ = jnp.zeros_like(T_e)
  for ion_symbol, fraction in zip(ion_symbols, ion_mixture.fractions):
    effective_LZ += fraction * _calculate_impurity_radiation_single_species(
        T_e, ion_symbol
    )
  return effective_LZ


def impurity_radiation_mavrin_fit(
    static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
    dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
    unused_geo: geometry.Geometry,
    source_name: str,
    core_profiles: state.CoreProfiles,
    unused_calculated_source_profiles: source_profiles.SourceProfiles | None,
    unused_conductivity: conductivity_base.Conductivity | None,
) -> tuple[chex.Array, ...]:
  """Model function for impurity radiation heat sink."""
  effective_LZ = calculate_total_impurity_radiation(
      ion_symbols=static_runtime_params_slice.impurity_names,
      ion_mixture=dynamic_runtime_params_slice.plasma_composition.impurity,
      T_e=core_profiles.T_e.value,
  )
  dynamic_source_runtime_params = dynamic_runtime_params_slice.sources[
      source_name
  ]
  assert isinstance(dynamic_source_runtime_params, DynamicRuntimeParams)
  radiation_profile = (
      effective_LZ
      * core_profiles.n_e.value
      * core_profiles.n_impurity.value
      * dynamic_source_runtime_params.radiation_multiplier
      * dynamic_runtime_params_slice.numerics.density_reference**2
  )

  # The impurity radiation heat sink is a negative source, so we return a
  # negative profile.
  return (-radiation_profile,)


class ImpurityRadiationHeatSinkMavrinFitConfig(base.SourceModelBase):
  """Configuration for the ImpurityRadiationHeatSink.

  Attributes:
    radiation_multiplier: Multiplier for the impurity radiation profile.
  """

  model_name: Literal['mavrin_fit'] = 'mavrin_fit'
  radiation_multiplier: float = 1.0
  mode: runtime_params_lib.Mode = runtime_params_lib.Mode.MODEL_BASED

  @property
  def model_func(self) -> source_lib.SourceProfileFunction:
    return impurity_radiation_mavrin_fit

  def build_dynamic_params(
      self,
      t: chex.Numeric,
  ) -> 'DynamicRuntimeParams':
    return DynamicRuntimeParams(
        prescribed_values=tuple(
            [v.get_value(t) for v in self.prescribed_values]
        ),
        radiation_multiplier=self.radiation_multiplier,
    )

  def build_source(
      self,
  ) -> impurity_radiation_heat_sink.ImpurityRadiationHeatSink:
    return impurity_radiation_heat_sink.ImpurityRadiationHeatSink(
        model_func=self.model_func
    )


@chex.dataclass(frozen=True)
class DynamicRuntimeParams(runtime_params_lib.DynamicRuntimeParams):
  radiation_multiplier: array_typing.ScalarFloat



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class for bootstrap current models."""
import abc

import chex
import jax
import jax.numpy as jnp
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry as geometry_lib
from torax._src.neoclassical.bootstrap_current import runtime_params
from torax._src.torax_pydantic import torax_pydantic

# pylint: disable=invalid-name


@chex.dataclass(kw_only=True, frozen=True)
class BootstrapCurrent:
  """Values returned by a bootstrap current model."""
  j_bootstrap: jax.Array
  j_bootstrap_face: jax.Array

  @classmethod
  def zeros(cls, geometry: geometry_lib.Geometry) -> 'BootstrapCurrent':
    """Returns a BootstrapCurrent with all values set to zero."""
    return cls(
        j_bootstrap=jnp.zeros_like(geometry.rho_norm),
        j_bootstrap_face=jnp.zeros_like(geometry.rho_face_norm),
    )


class BootstrapCurrentModel(abc.ABC):
  """Base class for bootstrap current models."""

  @abc.abstractmethod
  def calculate_bootstrap_current(
      self,
      dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
      geometry: geometry_lib.Geometry,
      core_profiles: state.CoreProfiles,
  ) -> BootstrapCurrent:
    """Calculates bootstrap current."""


class BootstrapCurrentModelConfig(torax_pydantic.BaseModelFrozen, abc.ABC):
  """Base class for bootstrap current model configs."""

  @abc.abstractmethod
  def build_dynamic_params(self) -> runtime_params.DynamicRuntimeParams:
    """Builds dynamic runtime params."""

  @abc.abstractmethod
  def build_model(self) -> BootstrapCurrentModel:
    """Builds bootstrap current model."""



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The LinearThetaMethod solver class."""
import jax
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.fvm import calc_coeffs
from torax._src.fvm import cell_variable
from torax._src.geometry import geometry
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.solver import predictor_corrector_method
from torax._src.solver import solver as solver_lib
from torax._src.sources import source_profiles


class LinearThetaMethod(solver_lib.Solver):
  """Time step update using theta method, linearized on coefficients at t."""

  def _x_new(
      self,
      dt: jax.Array,
      static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
      dynamic_runtime_params_slice_t: runtime_params_slice.DynamicRuntimeParamsSlice,
      dynamic_runtime_params_slice_t_plus_dt: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo_t: geometry.Geometry,
      geo_t_plus_dt: geometry.Geometry,
      core_profiles_t: state.CoreProfiles,
      core_profiles_t_plus_dt: state.CoreProfiles,
      core_sources_t: source_profiles.SourceProfiles,
      core_transport_t: state.CoreTransport,
      explicit_source_profiles: source_profiles.SourceProfiles,
      evolving_names: tuple[str, ...],
  ) -> tuple[
      tuple[cell_variable.CellVariable, ...],
      source_profiles.SourceProfiles,
      conductivity_base.Conductivity,
      state.CoreTransport,
      state.SolverNumericOutputs,
  ]:
    """See Solver._x_new docstring."""

    # Not used in this implementation.
    del core_sources_t, core_transport_t

    x_old = tuple([core_profiles_t[name] for name in evolving_names])
    x_new_guess = tuple(
        [core_profiles_t_plus_dt[name] for name in evolving_names]
    )

    coeffs_callback = calc_coeffs.CoeffsCallback(
        static_runtime_params_slice=static_runtime_params_slice,
        transport_model=self.transport_model,
        explicit_source_profiles=explicit_source_profiles,
        source_models=self.source_models,
        pedestal_model=self.pedestal_model,
        evolving_names=evolving_names,
    )

    # Compute the explicit coeffs based on the core profiles at time t and all
    # runtime parameters at time t.
    coeffs_exp = coeffs_callback(
        dynamic_runtime_params_slice_t,
        geo_t,
        core_profiles_t,
        x_old,
        allow_pereverzev=True,
        explicit_call=True,
    )

    # Calculate x_new with the predictor corrector method. Reverts to a
    # standard linear solve if
    # static_runtime_params_slice.predictor_corrector=False.
    # init_val is the initialization for the predictor_corrector loop.
    x_new = predictor_corrector_method.predictor_corrector_method(
        dt=dt,
        static_runtime_params_slice=static_runtime_params_slice,
        dynamic_runtime_params_slice_t_plus_dt=(
            dynamic_runtime_params_slice_t_plus_dt
        ),
        geo_t_plus_dt=geo_t_plus_dt,
        x_old=x_old,
        x_new_guess=x_new_guess,
        core_profiles_t_plus_dt=core_profiles_t_plus_dt,
        coeffs_exp=coeffs_exp,
        coeffs_callback=coeffs_callback,
    )

    coeffs_final = coeffs_callback(
        dynamic_runtime_params_slice_t_plus_dt,
        geo_t_plus_dt,
        core_profiles_t_plus_dt,
        x_new,
        allow_pereverzev=True,
    )
    core_sources, core_conductivity, core_transport = (
        coeffs_final.auxiliary_outputs
    )

    solver_numeric_outputs = state.SolverNumericOutputs(
        inner_solver_iterations=1,
        solver_error_state=0,  # linear method always works
    )

    return (
        x_new,
        core_sources,
        core_conductivity,
        core_transport,
        solver_numeric_outputs,
    )



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The ConstantTransportModel class.

A simple model assuming constant transport.
"""
import chex
from jax import numpy as jnp
from torax._src import array_typing
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.pedestal_model import pedestal_model as pedestal_model_lib
from torax._src.transport_model import runtime_params as runtime_params_lib
from torax._src.transport_model import transport_model


# pylint: disable=invalid-name
@chex.dataclass(frozen=True)
class DynamicRuntimeParams(runtime_params_lib.DynamicRuntimeParams):
  """Extends the base runtime params with additional params for this model.

  See base class runtime_params.DynamicRuntimeParams docstring for more info.
  """

  chi_i: array_typing.ScalarFloat
  chi_e: array_typing.ScalarFloat
  D_e: array_typing.ScalarFloat
  V_e: array_typing.ScalarFloat


class ConstantTransportModel(transport_model.TransportModel):
  """Calculates various coefficients related to particle transport."""

  def __init__(self):
    super().__init__()
    self._frozen = True

  def _call_implementation(
      self,
      dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
      geo: geometry.Geometry,
      core_profiles: state.CoreProfiles,
      pedestal_model_outputs: pedestal_model_lib.PedestalModelOutput,
  ) -> state.CoreTransport:
    del (
        core_profiles,
        pedestal_model_outputs,
    )  # Not needed for this transport model

    assert isinstance(
        dynamic_runtime_params_slice.transport, DynamicRuntimeParams
    )

    chi_face_ion = (
        dynamic_runtime_params_slice.transport.chi_i
        * jnp.ones_like(geo.rho_face)
    )
    chi_face_el = (
        dynamic_runtime_params_slice.transport.chi_e
        * jnp.ones_like(geo.rho_face)
    )
    d_face_el = dynamic_runtime_params_slice.transport.D_e * jnp.ones_like(
        geo.rho_face
    )
    v_face_el = dynamic_runtime_params_slice.transport.V_e * jnp.ones_like(
        geo.rho_face
    )

    return state.CoreTransport(
        chi_face_ion=chi_face_ion,
        chi_face_el=chi_face_el,
        d_face_el=d_face_el,
        v_face_el=v_face_el,
    )

  def __hash__(self):
    # All ConstantTransportModels are equivalent and can hash the same
    return hash('ConstantTransportModel')

  def __eq__(self, other):
    # All ConstantTransportModels are equivalent
    return isinstance(other, ConstantTransportModel)



# Copyright 2024 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Pellet source for the n_e equation."""
import dataclasses
from typing import ClassVar, Literal

import chex
from torax._src import array_typing
from torax._src import state
from torax._src.config import runtime_params_slice
from torax._src.geometry import geometry
from torax._src.neoclassical.conductivity import base as conductivity_base
from torax._src.sources import base
from torax._src.sources import formulas
from torax._src.sources import runtime_params as runtime_params_lib
from torax._src.sources import source
from torax._src.sources import source_profiles
from torax._src.torax_pydantic import torax_pydantic


# Default value for the model function to be used for the pellet source
# source. This is also used as an identifier for the model function in
# the default source config for Pydantic to "discriminate" against.
DEFAULT_MODEL_FUNCTION_NAME: str = 'gaussian'


# pylint: disable=invalid-name
def calc_pellet_source(
    unused_static_runtime_params_slice: runtime_params_slice.StaticRuntimeParamsSlice,
    dynamic_runtime_params_slice: runtime_params_slice.DynamicRuntimeParamsSlice,
    geo: geometry.Geometry,
    source_name: str,
    unused_state: state.CoreProfiles,
    unused_calculated_source_profiles: source_profiles.SourceProfiles | None,
    unused_conductivity: conductivity_base.Conductivity | None,
) -> tuple[chex.Array, ...]:
  """Calculates external source term for n from pellets."""
  dynamic_source_runtime_params = dynamic_runtime_params_slice.sources[
      source_name
  ]
  assert isinstance(dynamic_source_runtime_params, DynamicPelletRuntimeParams)
  return (
      formulas.gaussian_profile(
          center=dynamic_source_runtime_params.pellet_deposition_location,
          width=dynamic_source_runtime_params.pellet_width,
          total=(
              dynamic_source_runtime_params.S_total
              / dynamic_runtime_params_slice.numerics.density_reference
          ),
          geo=geo,
      ),
  )


@dataclasses.dataclass(kw_only=True, frozen=True, eq=True)
class PelletSource(source.Source):
  """Pellet source for the n_e equation."""

  SOURCE_NAME: ClassVar[str] = 'pellet'
  model_func: source.SourceProfileFunction = calc_pellet_source

  @property
  def source_name(self) -> str:
    return self.SOURCE_NAME

  @property
  def affected_core_profiles(self) -> tuple[source.AffectedCoreProfile, ...]:
    return (source.AffectedCoreProfile.NE,)


@chex.dataclass(frozen=True)
class DynamicPelletRuntimeParams(runtime_params_lib.DynamicRuntimeParams):
  pellet_width: array_typing.ScalarFloat
  pellet_deposition_location: array_typing.ScalarFloat
  S_total: array_typing.ScalarFloat


class PelletSourceConfig(base.SourceModelBase):
  """Pellet source for the n_e equation.

  Attributes:
    pellet_width: Gaussian width of pellet deposition [normalized radial coord]
    pellet_deposition_location: Gaussian center of pellet deposition [normalized
      radial coord]
    S_total: total pellet particles/s
    mode: Defines how the source values are computed (from a model, from a file,
      etc.)
  """

  model_name: Literal['gaussian'] = 'gaussian'
  pellet_width: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.1)
  )
  pellet_deposition_location: torax_pydantic.TimeVaryingScalar = (
      torax_pydantic.ValidatedDefault(0.85)
  )
  S_total: torax_pydantic.TimeVaryingScalar = torax_pydantic.ValidatedDefault(
      2e22
  )
  mode: runtime_params_lib.Mode = runtime_params_lib.Mode.MODEL_BASED

  @property
  def model_func(self) -> source.SourceProfileFunction:
    return calc_pellet_source

  def build_dynamic_params(
      self,
      t: chex.Numeric,
  ) -> DynamicPelletRuntimeParams:
    return DynamicPelletRuntimeParams(
        prescribed_values=tuple(
            [v.get_value(t) for v in self.prescribed_values]
        ),
        pellet_width=self.pellet_width.get_value(t),
        pellet_deposition_location=self.pellet_deposition_location.get_value(t),
        S_total=self.S_total.get_value(t),
    )

  def build_source(self) -> PelletSource:
    return PelletSource(model_func=self.model_func)
